/*
 *  Project: Replica exchange Monte Carlo for Protein Folding in the HP model
 *  File:    Conformation.cpp
 *
 *  Note: The latest source code and information regarding this project 
 *  is available at http://www.cs.ubc.ca/labs/beta/Projects/REMC-HPPFP/
 *
 *  For citation purposes, please use:
 *  Thachuk C, Shmygelska A and Hoos H: Replica exchange Monte Carlo for 
 *  protein folding in the HP model. BMC Bioinformatics 2007, Submitted.
 * 
 *  Copyright (C) 2007  Chris Thachuk, Alena Shmygelska and Holger H Hoos
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include "Conformation.h"

int Conformation::OFFSET=0;
double Conformation::probAcceptValues[MAX_NUM_REPLICAS][MAX_NUM_ENERGY_VALUES];
double Conformation::Tvalues[MAX_NUM_REPLICAS];

/*************************************
/* Public Interface
/************************************/

// default constructor
Conformation::Conformation(const string &seq, const double t, const string &expID, const int repID)
	:searchResult(expID, repID) 
{
	setSequence( seq );
	
	replicaID = repID;
	i0 = I0; //(HashSize/2-L*seqLength/2+L/2);
		
	for(int i=0; i<HashSize; i++) {
		Lattice[i] = LPempty;				Lattice_M[i] = LPempty;
	}
	
	// lay out the sequence initially in a straight
	// line conformation
	int residueType;
	for(int i=0; i<seqLength; i++) {
		if( sequence[i] == P ) {
			Lattice[i0 + (i - (int)(seqLength/2))*L] = P; 	Lattice_M[i0 + (i - (int)(seqLength/2))*L] = P; 
			sequenceCode[i] = P;
		} else {
			Lattice[i0 + (i - (int)(seqLength/2))*L] = i; 	Lattice_M[i0 + (i - (int)(seqLength/2))*L] = i; 
			sequenceCode[i] = i;
		}
		ii[i] = i0 + (i - (int)(seqLength/2))*L; 	ii_M[i] = i0 + (i - (int)(seqLength/2))*L;
		Directions[i] = straight;
	}
	
	// set first two residues as the grounded residues
	Directions[0] = ground;
	Directions[1] = ground;
	
	FixedPoint = 2;
	FreeEnergy = 0; FreeEnergy_M = 0;	
	T = t;
	prevTime = 0;
	prevMCStep = 0;
	pmWeight = SearchParameters::pmWeight;
}

// informs SearchResult to dump results to file
void Conformation::writeResults() {
	searchResult.writeResults();
}

// returns the current temperature of this conformation
double Conformation::getTemperature() {
	return Tvalues[replicaID];
}

// sets the temperature of this conformation
void Conformation::setTemperature(double t) {
	T = t;
}

// returns the current free energy of this conformation
int Conformation::getFreeEnergy() {
	return FreeEnergy;
}

// returns the replica id
int Conformation::getReplicaID() {
	return replicaID;
}

SearchResult& Conformation::getSearchResult() {
	return searchResult;
}

// runs the MC simulation on this conformation for n steps
SearchResult& Conformation::runSimulation(long double n) {
	int mutationPoint;

	acceptedMoves = 0;
	rejectedMoves = 0;
	impossibleMoves = 0;
	numResidues = 0;
	long double i;

	for(i = 0; i < n; i++) {
		mutationPoint = random_number(0,seqLength-1);

#if MOVESET == LOCAL	
		if( mutationPoint == 0 || mutationPoint == seqLength-1 ) {
			tryEndMove( mutationPoint );
			
//			cerr << "End Move" << endl;
		} else {
//#if DIM == 3
			double Q = ran01();
			if( Q >= 0.8 ) {
				tryCrankshaftMove( mutationPoint );
				//				cerr << "Crankshaft" << endl;
				
			} else {
				tryCornerMove( mutationPoint );
				//				cerr << "Corner" << endl;
			}
//#elif DIM == 2
//			tryCrankshaftOrCornerMove( mutationPoint );
//#endif
		}

#elif MOVESET == PULLMOVES
		if( mutationPoint == 0 || mutationPoint == seqLength-1 ) {
			tryEndPullMove( mutationPoint );
		} else {
			tryPullMove( mutationPoint );
		}

#elif MOVESET == MIXED
			
		if( mutationPoint == 0 || mutationPoint == seqLength-1 ) {
			double Q = ran01();
			if( Q <= pmWeight ) {
				tryEndPullMove( mutationPoint );
			} else {
				tryEndMove( mutationPoint );
			}
		} else {
			double Q = ran01();
			if( Q <= pmWeight ) {
				tryPullMove( mutationPoint );
			} else {
//#if DIM == 2
//				tryCrankshaftOrCornerMove( mutationPoint );
//#elif DIM == 3
				double Q = ran01();
				if( Q >= 0.8 ) {
					tryCrankshaftMove( mutationPoint );
				} else {
					tryCornerMove( mutationPoint );
				}
//#endif			
			}
		}
#endif
		
		if( move_isPossible ) {
			
			int energyDelta = FreeEnergy_M - FreeEnergy;
			
			if( energyDelta < 0 ) {
//				int oldEnergy = FreeEnergy;
//				printRelLatticePositions();
				applyMutation();
//				printRelLatticePositions();
//				if(!testFreeEnergy(oldEnergy+energyDelta)) {
//					cout << "Problem" << endl;
//				}
//				printRelDirections();
//				printRelLatticePositions();
				//assert(testLatticePositions());
				//testFreeEnergy(FreeEnergy);
				acceptedMoves++;
			} else {
				//double beta = 1.0 / (K_b*T);
				double prob = probAcceptValues[replicaID][energyDelta]; //exp(-energyDelta/(T*K_b));
			
				double Q = ran01();
			
				if( prob >= Q /*Q >=0.5*/ ) {
//					int oldEnergy = FreeEnergy;
//					printRelLatticePositions();
					applyMutation();
//					printRelLatticePositions();
//					if(!testFreeEnergy(oldEnergy+energyDelta)) {
//						cout << "Problem" << endl;
//					}
//					printRelDirections();
//					printRelLatticePositions();
					//assert(testLatticePositions());
					//testFreeEnergy(FreeEnergy);
					acceptedMoves++;
				} else {
					rejectedMoves++;
					resetMutation();
				}
			}
			
		} else {
			impossibleMoves++;
		}
		
		if( FreeEnergy < bestFreeEnergy ) {
			recalculateDirections();
			searchResult.addSQT_item(prevMCStep + i, 0 /*prevTime + t_clock.elapsed()*/, 
				FreeEnergy, Tvalues[replicaID], Directions, seqLength);
			bestFreeEnergy = FreeEnergy;
			for(int i=0; i<seqLength; i++)
				bestRelDirections[i] = Directions[i];
		}
		   
	}
	
	//t_clock.stop();
	//prevTime += t_clock.elapsed();
	prevMCStep += n - 1;
	
	searchResult.setAcceptedMoves(acceptedMoves);
	searchResult.setImpossibleMoves(impossibleMoves);
	searchResult.setRejectedMoves(rejectedMoves);
	searchResult.setNumResidues(numResidues);
	searchResult.setNumSteps(i-1);

	return searchResult;
}

void Conformation::recalculateDirections() {
	int GBasis[DIM][DIM];
	
	Directions[0] = ground;
    Directions[1] = ground;
	determineBasis(0, 1, GBasis);
	
	for(int i=2; i<seqLength; i++) {
		determineDirection(i, i-1, i, Directions[i], GBasis, 2);
	}
	//testLatticePositions();
}

/*************************************
/* Private Methods
/************************************/

#if MOVESET == LOCAL || MOVESET == MIXED
 // attempts to perfom an end move mutation
 // returns the energy differential if the move were to take place
 // sets the mutation m to the moves necessary to make the mutation
 void Conformation::tryEndMove(int residue) {
	//mutation.clear();
	//cout << "Begin End Move" << endl;
	int freeNeighbours[2*DIM];
	int bondedResidue;
	int plp, nlp;
	move_isPossible = false;
	
	if( residue == 0 ) {
		bondedResidue = 1;
		plp = -1;
		nlp = ii[residue + 1];
	}
	else if(residue == (seqLength-1)) {
		bondedResidue = seqLength-2;
		plp = ii[residue - 1];
		nlp = -1;
	}
	else {
		return; // mutation;
	}
	
	int num = findFreeNeighbours(bondedResidue, freeNeighbours);
		
	// if no free positions, simply return
	if(num == 0) {
		return; // mutation;
	}
		
	// select one of the free positions at random
	int lp = selectRandomFreeNeighbour(freeNeighbours, num);
	Lattice_M[ii_M[residue]] = LPempty;
	Lattice_M[lp] = sequenceCode[residue];
	ii_M[residue] = lp;
	
	// determine the energy differential if we move the residue here
	int bondsBefore = countBonds(residue, ii[residue], plp, nlp);
	int bondsAfter = countBonds(residue, lp, plp, nlp); 
	
	
	FreeEnergy_M += bondsBefore - bondsAfter;
	move_startResidue = residue;
	move_endResidue = residue;
	move_isPossible = true;
	
	//cout << "End End Move" << endl;
	return; // mutation;
}


//void Conformation::tryCrankshaftOrCornerMove(int residue) {
//	move_isPossible = false;
//	//mutation.clear();
//	int n1=-1, n2=-1;
//	//cout << "Begin Crankshaft" << endl;
//	
//	// can't do a corner or crankshaft move on first or last residue
//	if( residue == 0 || residue == seqLength-1 ) {
//		return; // mutation;
//	}
//	
//	//if( residue > 1 && residue < seqLength-2 ) {
//	if ( Directions[residue] == ground ) {
//		
//		// check for [G]RRL case where residue is the G
//		if( Directions[residue+1]==right && Directions[residue+2]==right && Directions[residue+3]==left ) {
//			//cout << "[G]RRL" << endl;
//			n1 = residue;
//			n2 = residue+1;
//		}
//		// check for [G]LLR case where residue is the G
//		else if( Directions[residue+1]==left && Directions[residue+2]==left && Directions[residue+3]==right ) {
//			//cout << "[G]LLR" << endl;
//			n1 = residue;
//			n2 = residue+1;		
//		}
//		
//	}
//	// normal case in middle of sequence where residue direction is R
//	else if( Directions[residue] == right ) {
//		
//		// check for G[R]RL case where residue is the first R
//		if ( Directions[residue-1]==ground && Directions[residue+1]==right && Directions[residue+2]==left ) {
//			//cout << "G[R]RL" << endl;
//			n1 = residue-1;
//			n2 = residue;
//		}
//		// check for [R]LL$ case where residue is the first R and $ is the end of sequence
//		else if( residue==seqLength-3 && Directions[residue+1]==left && Directions[residue+2]==left ) {
//			//cout << "[R]LL$" << endl;
//			n1 = residue;
//			n2 = residue+1;
//		}
//		// check for L[R]R$ case where residue is the first R and $ is the end of sequence
//		else if( residue==seqLength-2 && Directions[residue-1]==left && Directions[residue+1]==right ) {
//			//cout << "L[R]R$" << endl;
//			n1 = residue-1;
//			n2 = residue;
//		}
//		// check for [R]LLR case where residue is the first R
//		else if( (residue<seqLength-3) && Directions[residue+1]==left && Directions[residue+2]==left && Directions[residue+3]==right ) {
//			//cout << "[R]LLR" << endl;		
//			n1 = residue;
//			n2 = residue+1;
//		} 
//		// check for L[R]RL case where residue is the first R
//		else if( (residue<seqLength-2) && Directions[residue-1]==left && Directions[residue+1]==right && Directions[residue+2]==left ) {
//			//cout << "L[R]RL" << endl;
//			n1 = residue-1;
//			n2 = residue;
//		}
//	
//	} 
//	// normal case in middle of sequence where residue direction is L
//	else if( Directions[residue] == left ) {
//	
//		// check for G[L]LR case where residue is the first L
//		if( Directions[residue-1]==ground && Directions[residue+1]==left && Directions[residue+2]==right ) {
//			//cout << "G[L]LR" << endl;
//			n1 = residue-1;
//			n2 = residue;
//		}
//		// check for [L]RR$ case where residue is the first L and $ is the end of sequence
//		else if( residue==seqLength-3 && Directions[residue+1]==right && Directions[residue+2]==right ) {
//			//cout << "[L]RR$" << endl;
//			n1 = residue;
//			n2 = residue+1;
//		}
//		// check for R[L]L$ case where residue is the first L and $ is the end of sequence
//		else if( residue==seqLength-2 && Directions[residue-1]==right && Directions[residue+1]==left ) {
//			//cout << "R[L]L$" << endl;
//			n1 = residue-1;
//			n2 = residue;
//		}
//		// check for [L]RRL case where residue is the first L
//		else if( (residue<seqLength-3) && Directions[residue+1]==right && Directions[residue+2]==right && Directions[residue+3]== left ) {
//			//cout << "[L]RRL" << endl;		
//			n1 = residue;
//			n2 = residue+1;
//		}
//		// check for R[L]LR case where residue is the first L
//		else if( (residue<seqLength-2) && Directions[residue-1]==right && Directions[residue+1]==left && Directions[residue+2]==right ) {
//			//cout << "R[L]LR" << endl;	
//			n1 = residue-1;
//			n2 = residue;
//		}
//		
//	}
//	//}
//	
//	// crankshaft recognized, see if the opposing positions are free for the move
//	if( n1 != -1 ) {
//		int new_n1 = 2*ii[n1-1]-ii[n1];
//		int new_n2 = 2*ii[n2+1]-ii[n2];
//		//cout << residue << "," << n1 << "," << n2 << "," << new_n1 << "," << new_n2 << endl;
//		//printLatticePositions();
//		//printRelDirections();
//	
//		if( Lattice[new_n1]==LPempty && Lattice[new_n2]==LPempty ) {
//		
//			// determine the energy differential if we move the residues here
//			// residue1
//			int bondsBefore = countBonds(n1, ii[n1], ii[n1-1], ii[n1+1]);
//			int bondsAfter = countBonds(n1, new_n1, ii[n1-1], new_n2); 
//			//int energyDelta = bondsBefore - bondsAfter;
//			FreeEnergy_M += bondsBefore - bondsAfter;
//			
//			// residue2
//			bondsBefore = countBonds(n2, ii[n2], ii[n2-1], ii[n2+1]);
//			bondsAfter = countBonds(n2, new_n2, new_n1, ii[n2+1]);
//			//energyDelta += (bondsBefore - bondsAfter);
//			FreeEnergy_M += bondsBefore - bondsAfter;
//	
//			//mutation.setEnergyDelta(energyDelta);
//			//mutation.push(n1, new_n1);
//			//mutation.push(n2, new_n2);
//			
//			Lattice_M[ii[n1]] = LPempty;
//			Lattice_M[ii[n2]] = LPempty;
//			
//			ii_M[n1] = new_n1;
//			ii_M[n2] = new_n2;
//			
//			Lattice_M[new_n1] = sequenceCode[n1];
//			Lattice_M[new_n2] = sequenceCode[n2];
//			
//			if(n1 > n2) {
//				int tmp = n1;
//				n1 = n2;
//				n2 = tmp;
//			}
//			move_startResidue = n1;
//			move_endResidue = n2;
//			move_isPossible = true;
//		}
//		//cout << "Done Crankshaft" << endl;
//	}
//	// this residue is not part of a crankshaft, see if it's a corner
//	else if( Directions[residue+1] != straight ) {
//		//cout << "End Crankshaft" << endl;
//		//cout << "Begin Corner" << endl;
//		// check union of free neighbours for residue-1 and residue+1
//		// there is only 1 possibility for corner move
//		int freeN[DIM*2];
//		int numFree;
//		
//		if( (numFree=findFreeNeighbours( residue-1, residue+1, freeN )) > 0 ) {
//			int new_n1 = selectRandomFreeNeighbour(freeN, numFree);
//			// determine the energy differential if we move the residue here
//			int bondsBefore = countBonds(residue, ii[residue], ii[residue-1], ii[residue+1]);
//			int bondsAfter = countBonds(residue, new_n1, ii[residue-1], ii[residue+1]); 
//			//int energyDelta = bondsBefore - bondsAfter;
//			FreeEnergy_M += bondsBefore - bondsAfter;
//			//mutation.setEnergyDelta(energyDelta);
//			Lattice_M[ii[residue]] = LPempty;
//			ii_M[residue] = new_n1;
//			Lattice_M[new_n1] = sequenceCode[residue];
//			move_startResidue = residue;
//			move_endResidue = residue;
//			move_isPossible = true;
//			//mutation.push(residue,new_n1);
//		}
//		//cout << "End Corner" << endl;
//	}
//	
//	return;
//}
//

void Conformation::tryCrankshaftMove(int residue) {
#if DIM == 2
	static const int componentSum = (1 + L);
#elif DIM == 3
	static const int componentSum = (1 + L + LL);
#endif

	int newNiplus1 = -1, newNiplus2 = -1;
	bool canCrankshaft = false;
	
	move_isPossible = false;
		
	//cout << "Begin Crankshaft" << endl;
	
	// can't do a corner or crankshaft move on first or last residue
	if( residue == 0 || residue == seqLength-1 ) {
		return;
	}
	
	int abs_i_to_i_plus_3;
	
	if( residue + 3 < seqLength - 1 ) {
		abs_i_to_i_plus_3 = abs(ii[residue+3] - ii[residue]);
		
#if DIM == 2
		if( abs_i_to_i_plus_3 == 1 || abs_i_to_i_plus_3 == L )
#elif DIM == 3
		if( abs_i_to_i_plus_3 == 1 || abs_i_to_i_plus_3 == L || abs_i_to_i_plus_3 == LL  )
#endif
			canCrankshaft = true;
	}
	
	if( !canCrankshaft && residue - 3 >= 1 ) {
		residue = residue - 3;
		abs_i_to_i_plus_3 = abs(ii[residue+3] - ii[residue]);

#if DIM == 2
		if( abs_i_to_i_plus_3 == 1 || abs_i_to_i_plus_3 == L )
#elif DIM == 3
		if( abs_i_to_i_plus_3 == 1 || abs_i_to_i_plus_3 == L || abs_i_to_i_plus_3 == LL  )
#endif
			canCrankshaft = true;
	}	
	
	if( canCrankshaft ) {
		int abs_i_to_i_plus_1 = abs(ii[residue+1] - ii[residue]);
		int distance = componentSum - abs_i_to_i_plus_3 - abs_i_to_i_plus_1;
			
		// randomly select which crankshaft direction to try first
		int direction = random_number(1,2);
		if( direction == 2 ) distance = -distance;
			
		if( (Lattice[ii[residue] + distance] == LPempty) && (Lattice[ii[residue + 3] + distance] == LPempty) ) {
			newNiplus1 = ii[residue] + distance;
			newNiplus2 = ii[residue + 3] + distance; 
		} else if( (Lattice[ii[residue] - distance] == LPempty) && (Lattice[ii[residue + 3] - distance] == LPempty) ) {
			newNiplus1 = ii[residue] - distance;
			newNiplus2 = ii[residue + 3] - distance;
		}
			
		if( newNiplus1 != -1 && newNiplus2 != -1 ) {
			// determine the energy differential if we move the residues here
			// residue + 1
			int bondsBefore = countBonds(residue + 1, ii[residue+1], ii[residue], ii[residue+2]);
			int bondsAfter = countBonds(residue + 1, newNiplus1, ii[residue], newNiplus2); 
			FreeEnergy_M += bondsBefore - bondsAfter;
			// residue + 2
			bondsBefore = countBonds(residue + 2, ii[residue+2], ii[residue+1], ii[residue+3]);
			bondsAfter = countBonds(residue + 2, newNiplus2, newNiplus1, ii[residue+3]);
			FreeEnergy_M += bondsBefore - bondsAfter;
			
			int n1=residue+1;
			int n2=residue+2;
			
			Lattice_M[ii[n1]] = LPempty;
			Lattice_M[ii[n2]] = LPempty;
			
			ii_M[n1] = newNiplus1;
			ii_M[n2] = newNiplus2;
			
			Lattice_M[newNiplus1] = sequenceCode[n1];
			Lattice_M[newNiplus2] = sequenceCode[n2];
			
			if(n1 > n2) {
				int tmp = n1;
				n1 = n2;
				n2 = tmp;
			}
			move_startResidue = n1;
			move_endResidue = n2;
			move_isPossible = true;
		}
	}
	return;
}

void Conformation::tryCornerMove(int residue) {
	move_isPossible = false;
	
	if( (ii[residue] - ii[residue-1]) != (ii[residue+1] - ii[residue]) ) {
		//cout << "End Crankshaft" << endl;
		//cout << "Begin Corner" << endl;
		// check union of free neighbours for residue-1 and residue+1
		// there is only 1 possibility for corner move
		int freeN[DIM*2];
		int numFree;
		
		if( (numFree=findFreeNeighbours( residue-1, residue+1, freeN )) > 0 ) {
			int new_n1 = selectRandomFreeNeighbour(freeN, numFree);
			// determine the energy differential if we move the residue here
			int bondsBefore = countBonds(residue, ii[residue], ii[residue-1], ii[residue+1]);
			int bondsAfter = countBonds(residue, new_n1, ii[residue-1], ii[residue+1]); 
			FreeEnergy_M += bondsBefore - bondsAfter;
			
			Lattice_M[ii[residue]] = LPempty;
			ii_M[residue] = new_n1;
			Lattice_M[new_n1] = sequenceCode[residue];
			
			move_startResidue = residue;
			move_endResidue = residue;
			move_isPossible = true;

		}
	}
	return;
}

#endif

#if MOVESET == PULLMOVES || MOVESET == MIXED

void Conformation::tryEndPullMove(int residue) {
	int bondsBefore = 0, bondsAfter = 0;
	int newLP1 = -1, newLP2 = -1;
	int startResidue = residue, endResidue = residue;
	int offset = 1;
	move_isPossible = false;
	
	if( residue == 0 ) {
		offset = -1;
	}
	
	int freeNeighbours[2*DIM];
	int numFree = findFreeNeighbours(residue, freeNeighbours);
	if( numFree < 1 ) return;
	
	newLP1 = selectRandomFreeNeighbour(freeNeighbours, numFree);
	
	numFree = findFreeNeighboursL(newLP1, freeNeighbours);
	if( numFree < 1 ) return;
	
	newLP2 = selectRandomFreeNeighbour(freeNeighbours, numFree);
	
	if( newLP2 == -1 ) return;
	
	int lastMovePos, lastPastMovePos, secondLastPastMovePos;
	int residueJ;
	
	// move the end residue
	Lattice_M[ii_M[residue]] = LPempty;
	Lattice_M[newLP2] = sequenceCode[residue];
	ii_M[residue] = newLP2;
	
	// move the chain neighbour of end reside
	Lattice_M[ii_M[residue - offset]] = LPempty;
	Lattice_M[newLP1] = sequenceCode[residue - offset];
	ii_M[residue - offset] = newLP1;
	endResidue = residue - offset;
	
	// we must continue to pull the chain until a 
	// valid configuration occurs
	secondLastPastMovePos = ii[residue];
	lastPastMovePos = ii[residue - offset];
	lastMovePos = newLP1;
	residueJ = residue - offset - offset;
	
	while( residueJ >= 0 && residueJ <= seqLength - 1 && 
#if DIM == 2
		   ( abs(ii[residueJ] - lastMovePos) != L && abs(ii[residueJ] - lastMovePos) != 1 ) ) {
#elif DIM == 3
		   ( abs(ii[residueJ] - lastMovePos) != L && abs(ii[residueJ] - lastMovePos) != 1 ) && (abs(ii[residueJ] - lastMovePos) != LL ) ) {
#endif
		
		endResidue = residueJ;
		Lattice_M[ii_M[residueJ]] = LPempty;
		ii_M[residueJ] = secondLastPastMovePos;
		Lattice_M[secondLastPastMovePos] = sequenceCode[residueJ];
						
		lastMovePos = secondLastPastMovePos;
		secondLastPastMovePos = lastPastMovePos;
		lastPastMovePos = ii[residueJ];
		residueJ = residueJ - offset;
	}
	
	if( startResidue > endResidue ) {
		int tmp = startResidue;
		startResidue = endResidue;
		endResidue = tmp;
	}
		
	for( int res = startResidue; res <= endResidue; res += 1 )
	{
		
		if( sequenceCode[res] >= 0 ) {
			int lp = ii[res];
			int olp;
			int oRes;
			
			for( int i = 0; i < 2*DIM; i += 1 )
			{
				
				olp = lp + DI[i];
				oRes = Lattice[olp];
				if( oRes >= 0 ) {
				
					if( abs(oRes - res) > 1 ) {
						bondsBefore++;
					
						if( oRes > res && oRes <= endResidue ) {
							bondsBefore--;
						}
					}	
		    	}
			}
			
			for( int i = 0; i < 2*DIM; i += 1 )
			{
				lp = ii_M[res];
				olp = lp + DI[i];
				oRes = Lattice_M[olp];
				if( oRes >= 0 ) {
				
					if( abs(oRes - res) > 1 ) {
						bondsAfter++;
					
						if( oRes > res && oRes <= endResidue ) {
							bondsAfter--;
						}
					}	
		    	}
			}
			
		}
	}
	
	FreeEnergy_M += bondsBefore - bondsAfter;
	move_startResidue = startResidue;
	move_endResidue = endResidue;
	move_isPossible = true;
}

void Conformation::tryPullMove(int residue) {
	int bondsBefore = 0, bondsAfter = 0;
	int offset = 1, Lpos = -1, Cpos = -1;
	int startResidue = residue, endResidue = residue;
	move_isPossible = false;
	
	// regular pull move

#if DIM == 2		
	int residue2, offset2;
	// randomly select which pull direction to try first
	int direction = random_number(1,2);
	if( direction == 1 ) offset = -offset;
	
	for(int i = 0; i < 2; i++) {
		residue2 = residue + offset;
	
		int d = abs(ii[residue] - ii[residue2]);
	
		if( d == L ) {
			offset2 = 1;
		} else {
		// d == 1
			offset2 = L;
		}
	
		// randomly select which direction to check first for L
		direction = random_number(1,2);
		if( direction == 1 ) {
			offset2 = -offset2;
		}
		
		if( Lattice[ii[residue2] + offset2] == LPempty 
			&& (Lattice[ii[residue] + offset2] == LPempty 
				|| ii[residue] + offset2 == ii[residue - offset] ) ) {
			
			Lpos = ii[residue2] + offset2;
			Cpos = ii[residue] + offset2;
			
		} else if( Lattice[ii[residue2] - offset2] == LPempty 
			&& (Lattice[ii[residue] - offset2] == LPempty
				|| ii[residue] - offset2 == ii[residue - offset] )) {
				
			Lpos = ii[residue2] - offset2;
			Cpos = ii[residue] - offset2;
		}
		
		if( Lpos != -1 ) break;
		
		// residue2 didn't have a free L location diagonal from residue1
		// so now select residue2 in the other direction (ie. reverse the pull move)
		offset = -offset;
	}

#elif DIM == 3
	int residue2, offset2, offset2b;
	// randomly select which pull direction to try first
	int direction = random_number(1,2);
	if( direction == 1 ) offset = -offset;
	
	for(int i = 0; i < 2; i++) {
		residue2 = residue + offset;
	
		int d = abs(ii[residue] - ii[residue2]);
	
		if( d == L ) {
			if( random_number(1,2) == 1 ) {
				offset2 = 1; offset2b = LL;
			} else {
				offset2 = LL; offset2b = 1;
			}
		} else if( d == LL ) {
			if( random_number(1,2) == 1 ) {
				offset2 = 1; offset2b = L;
			} else {
				offset2 = L; offset2b = 1;
			}
		} else {
			// d = 1
			if( random_number(1,2) == 1 ) {
				offset2 = L; offset2b = LL;
			} else {
				offset2 = LL; offset2b = L;
			}
		}
		
		for(int j=0; j<2; j++) {
	
			// randomly select which direction to check first for L
			direction = random_number(1,2);
			if( direction == 1 ) {
				offset2 = -offset2;
			}
		
			if( Lattice[ii[residue2] + offset2] == LPempty 
				&& (Lattice[ii[residue] + offset2] == LPempty 
					|| ii[residue] + offset2 == ii[residue - offset] ) ) {
			
				Lpos = ii[residue2] + offset2;
				Cpos = ii[residue] + offset2;
			
			} else if( Lattice[ii[residue2] - offset2] == LPempty 
				&& (Lattice[ii[residue] - offset2] == LPempty
					|| ii[residue] - offset2 == ii[residue - offset] )) {
				
				Lpos = ii[residue2] - offset2;
				Cpos = ii[residue] - offset2;
			}
		
			if( Lpos != -1 ) break;
			offset2 = offset2b;
		}
		
		if( Lpos != -1 ) break;
		// residue2 didn't have a free L location diagonal from residue1
		// so now select residue2 in the other direction (ie. reverse the pull move)
		offset = -offset;
	}

#endif
	
	// if we didn't find a free location L, simply return
	if( Lpos == -1 ) {
		return;
	}
	
	int lastMovePos, lastPastMovePos, secondLastPastMovePos;
	int residueJ;
	
	Lattice_M[ii_M[residue]] = LPempty;
	Lattice_M[Lpos] = sequenceCode[residue];
	ii_M[residue] = Lpos;
	
	// we now check if residue - offset is already in the C position,
	// if it is not, we must continue to pull the chain until a 
	// valid configuration occurs
	if( ii[residue - offset] != Cpos  ) {
		
		Lattice_M[ii_M[residue - offset]] = LPempty;
		ii_M[residue - offset] = Cpos;
		Lattice_M[Cpos] = sequenceCode[residue - offset];
		endResidue -= offset;
		
		secondLastPastMovePos = ii[residue];
		lastPastMovePos = ii[residue - offset];
		lastMovePos = Cpos;
		residueJ = residue - offset - offset;
		
		while( residueJ >= 0 && residueJ <= seqLength - 1 && 
#if DIM == 2
			   ( abs(ii[residueJ] - lastMovePos) != L && abs(ii[residueJ] - lastMovePos) != 1 ) ) {
#elif DIM == 3
			   ( abs(ii[residueJ] - lastMovePos) != L && abs(ii[residueJ] - lastMovePos) != 1 ) && abs(ii[residueJ] - lastMovePos) != LL ) {
#endif
			
			endResidue -= offset;
			Lattice_M[ii_M[residueJ]] = LPempty;
			ii_M[residueJ] = secondLastPastMovePos;
			Lattice_M[secondLastPastMovePos] = sequenceCode[residueJ];
							
			lastMovePos = secondLastPastMovePos;
			secondLastPastMovePos = lastPastMovePos;
			lastPastMovePos = ii[residueJ];
			residueJ = residueJ - offset;
		}
		
	}
	
	if( startResidue > endResidue ) {
		int tmp = startResidue;
		startResidue = endResidue;
		endResidue = tmp;
	}
		
	for( int res = startResidue; res <= endResidue; res += 1 )
	{
		
		if( sequenceCode[res] >= 0 ) {
			int lp = ii[res];
			int olp;
			int oRes;
			
			for( int i = 0; i < 2*DIM; i += 1 )
			{
				
				olp = lp + DI[i];
				oRes = Lattice[olp];
				if( oRes >= 0 ) {
				
					if( abs(oRes - res) > 1 ) {
						bondsBefore++;
					
						if( oRes > res && oRes <= endResidue ) {
							bondsBefore--;
						}
					}	
		    	}
			}
			
			for( int i = 0; i < 2*DIM; i += 1 )
			{
				lp = ii_M[res];
				olp = lp + DI[i];
				oRes = Lattice_M[olp];
				if( oRes >= 0 ) {
				
					if( abs(oRes - res) > 1 ) {
						bondsAfter++;
					
						if( oRes > res && oRes <= endResidue ) {
							bondsAfter--;
						}
					}	
		    	}
			}
			
		}
	}
	
	FreeEnergy_M += bondsBefore - bondsAfter;
	move_startResidue = startResidue;
	move_endResidue = endResidue;
	move_isPossible = true;
}

#endif

void Conformation::applyMutation() {
	int residue;
	int lp;
	int bondedResidue;
	int boundaryHit = false;
	
	// check if this is an empty mutation
	if( !move_isPossible ) return;	
	
	// let's clear the old lattice positions first
	for(residue=move_startResidue; residue <= move_endResidue; residue += 1) {
		Lattice[ii[residue]] = LPempty;
	}
	
	for(residue=move_startResidue; residue <= move_endResidue; residue += 1) {
		lp = ii_M[residue];
		//if( lp == 0) {
		//	cout << move_startResidue << "->" << move_endResidue << endl;
		//}
		int x = row(lp);
		int y = col(lp);
#if DIM == 3		
		int z = height(lp);
		
		if( x<=3 || y<=3 || z <= 3 || x>=L-3 || y>=L-3 || z>=L-3 ) {
			boundaryHit = true;
		}
#elif DIM == 2
		if( x<=3 || y<=3 || x>=L-3 || y>=L-3) {
			boundaryHit = true;
		}
#endif
		Lattice[lp] = sequenceCode[residue];
		ii[residue] = ii_M[residue];
	}

//#if DIM == 2	
//	// determine new basis matrix
//	int GBasis[DIM][DIM];
//		
//	if( move_startResidue == 0 || move_startResidue == 1) {
//		bondedResidue = 1;
//	} else {
//		bondedResidue = move_startResidue - 1;
//	}
//		
//	// always update relative direction of the bondedResidue + 1
//	determineBasis(bondedResidue - 1, bondedResidue, GBasis);
//	
//	while( bondedResidue <= move_endResidue + 2 && bondedResidue <= seqLength-2 ) {
//		determineDirection( bondedResidue + 1, bondedResidue, bondedResidue + 1, Directions[bondedResidue + 1], GBasis, 2);
//		bondedResidue++;
//	}
//#endif

	FreeEnergy = FreeEnergy_M;
	
	if( boundaryHit ) {
		recenterOnLattice();
	}
	//printConformation();
//	assert(testLatticePositions());
//	assert(testFreeEnergy(FreeEnergy));
	
	numResidues += move_endResidue - move_startResidue + 1;
	
}

void Conformation::resetMutation() {
	int residue;
	
	for(residue=move_startResidue; residue <= move_endResidue; residue += 1) {
		Lattice_M[ii_M[residue]] = LPempty;
	}
	
	for(residue=move_startResidue; residue <= move_endResidue; residue += 1) {
		ii_M[residue] = ii[residue];
		Lattice_M[ii_M[residue]] = sequenceCode[residue];
		FreeEnergy_M = FreeEnergy;
	}
}

// the lattice is always greater than 4 times the maximum
// sequence length in both width and height.  therefore, since
// this function is called only when a boundary is touched
// we can safely assume we can recenter the fold without overwriting
// the existing sequence, as we always translate the protein to the center
void Conformation:: recenterOnLattice() {

#ifdef TRACE_VERBOSE
	cout << endl;
	cout << "----------------------------------------------" << endl;
	cout << "| Recentering fold on Lattice (boundary hit) |" << endl;
	cout << "----------------------------------------------" << endl << endl;
	printLatticePositions();
#endif
	
	// step 1) erase Lattice
	for(int i = 0; i<seqLength; i++) {
		Lattice[ii[i]]=LPempty;
	}
	
	// step 2) Determine bounding box of embedding on Lattice_M
	int minRow = L + 1, minCol = L + 1, minHeight = L + 1, maxRow = 0, maxCol = 0, maxHeight = 0;
	
	for(int i=0; i<seqLength; i++) {
		int lrow = row(ii_M[i]);
		int lcol = col(ii_M[i]);
#if DIM == 3
		int lheight = height(ii_M[i]);
#endif
		
		if( lrow < minRow ) minRow = lrow;
		if( lrow > maxRow ) maxRow = lrow;
		if( lcol < minCol ) minCol = lcol;
		if( lcol > maxCol ) maxCol = lcol;
#if DIM == 3
		if( lheight < minHeight ) minHeight = lheight;
		if( lheight > maxHeight ) maxHeight = lheight;
#endif
	}
	
	// step 3) Determine the centroid of the embedding
	int r = minRow + (int)((maxRow - minRow)/2);
	int c = minCol + (int)((maxCol - minCol)/2);
#if DIM == 3
	int h = minHeight + (int)((maxHeight - minHeight)/2);
#endif
	
	int C = c + r * L;
#if DIM == 3
	C += h * LL;
#endif 
	
	int delta = i0 - C;


	
	// step 4) copy residues from Lattice_M to Lattice, translating by delta
	for(int i=0; i<seqLength; i++) {
		Lattice[ii_M[i]+delta]=Lattice_M[ii[i]];
		ii[i]=ii_M[i]+delta;
	}
	
	// step 5) erase Lattice_M
	for(int i=0; i<seqLength; i++) {
		Lattice_M[ii_M[i]] = LPempty;
	}
	
	// step 6) copy Lattice to Lattice_M and ii to ii_M
	for(int i=0; i<seqLength; i++) {
		Lattice_M[ii[i]] = Lattice[ii[i]];
		ii_M[i] = ii[i];
	}

#ifdef TRACE_VERBOSE	
	cout << endl << "Done translation move" << endl;
	printLatticePositions();
	cout << endl;
#endif
}

void Conformation:: printRelDirections() {
	for(int i=0; i<seqLength; i++)
		cout << Directions[i] << ",";
	cout << endl;
}

void Conformation:: printLatticePositions() {
	cout << "i0: (" << row(i0) << "," << col(i0) << ")" << endl;
	for(int i=0; i<seqLength; i++) {
		cout << "(" << row(ii[i]) << "," << col(ii[i]);
#if DIM == 3
		cout << "," << height(ii[i]);
#endif
		cout << ")" << ",";
	}
	cout << endl;
}

void Conformation:: printRelLatticePositions() {
#if DIM == 2
	cout << "(0,0)" << ",";
	cout << endl;
#elif DIM == 3
	cout << "(0,0,0)" << ",";
	cout << endl;
#endif
	for(int i=1; i<seqLength; i++) {
		cout << "(" << row(ii[i])-row(ii[0]) << ","
					<< col(ii[i])-col(ii[0]) 
#if DIM == 3					
					<< "," << height(ii[i])-height(ii[0]) 
#endif
					<< ")" << ",";
		cout << endl;
	}
	cout << endl;
}

bool Conformation:: testLatticePositions() {
	bool flag;
	
	flag=false;
	for(int j=0; j<2*DIM; j++) {
		if(ii[0] == ii[1]+DI[j] ) {
			flag=true;
			break;
		}
	}
	
	if( flag == false ) {
		cerr << "1. Something's wrong" << endl;
		return false;
		//throw "Something's wrong";
	}
	
	for(int i=1; i<seqLength; i++) {
		flag = false;
		for(int j=0; j<2*DIM; j++) {
			if( ii[i] == ii[i-1] + DI[j] ) {
				flag=true;
				break;
			}
		}
		if( flag==false ) {
			cerr << "2. Something's wrong" << endl;
			return false;
			//throw "Something's wrong";
		}
	}
	
	// now make sure no lattice points are shared by two residues
	for(int i = 0; i<seqLength-1; i++) {
		for(int j=i+1; j<seqLength; j++) {
			if(ii[i] == ii[j]) {
				cerr << "3. Something's wrong" << endl;
				return false;
			}
		}
	}
	
	return true;
}

int Conformation:: calcFreeEnergy() {
	int freeE = 0;
	
	freeE += countBonds(0,ii[0],-1,ii[1]);
	
	for(int i = 1; i<seqLength-1; i++) {
		freeE += countBonds(i,ii[i],ii[i-1],ii[i+1]);
	}
	
	freeE += countBonds(seqLength-1,ii[seqLength-1],ii[seqLength-2],-1);
	return -freeE/2;
}

bool Conformation:: testFreeEnergy(int E) {
	int realEnergy = calcFreeEnergy();
	
	if( realEnergy != E ) {
		cout << "Actual Free Energy: " << realEnergy << endl;
		cout << "Reported Free Energy: " << E << endl;
	}
	return realEnergy == E;
}

void Conformation::reset() {
        i0 = I0; //(HashSize/2-L*seqLength/2+L/2);
        //cout << "i0 (" << row(i0) << "," << col(i0) << ")" << endl;
        //cout << "I0 (" << row(I0) << "," << col(I0) << ")" << endl;
        
        for(int i=0; i<seqLength; i++) {
                Lattice[ii[i]] = LPempty;
				Lattice_M[ii_M[i]] = LPempty;
        }
        
        for(int i=0; i<seqLength; i++) {
            Directions[i] = straight;
			bestRelDirections[i] = straight;
        }
        
        FixedPoint = 2;
        FreeEnergy = 0;
		searchResult.resetMinimumSolution();
		bestFreeEnergy = 0;
		//mutation.clear();
}

void Conformation::makeStraight() {
	reset();
	// lay out the sequence initially in a straight
    // line conformation
    for(int i=0; i<seqLength; i++) {
            Lattice[i0 + i*L] = sequence[i];
			Lattice_M[i0 + i*L] = Lattice[i0 + i*L];
            ii[i] = i0 + i*L;
			ii_M[i] = ii[i];
            Directions[i] = straight;
    }

    // set first two residues as the grounded residues
    Directions[0] = ground;
    Directions[1] = ground;
}

void Conformation::computeFreeEnergy() {
	FreeEnergy = calcFreeEnergy();
}

string Conformation::getContactString() {
	ostringstream cs;
	bool firstContact = true;
	
	for(int i=0; i<seqLength; i++) {
		if( sequence[i] == P) continue;
		
		for(int j=i+2; j<seqLength; j++) {
			if( sequence[j] == P ) continue;
			
			int distance = abs(ii[i] - ii[j]);

#if DIM == 2			
			if( distance == 1 || distance == L ) {
#elif DIM == 3
			if( distance == 1 || distance == L || distance == LL ) {
#endif
				if(!firstContact) cs << ",";
				cs << "(" << (i+1) << "," << (j+1) << ")";
				firstContact = false;
			}
		}
	}
	
	return cs.str();
}

string Conformation::getSymmetricContactString() {
	ostringstream cs;
	bool firstContact = true;
	
	for(int i=seqLength-1; i>=0; i--) {
		if( sequence[i] == P) continue;
		
		for(int j=i-2; j>=0; j--) {
			if( sequence[j] == P ) continue;
			
			int distance = abs(ii[i] - ii[j]);

#if DIM == 2			
			if( distance == 1 || distance == L ) {
#elif DIM == 3
			if( distance == 1 || distance == L || distance == LL ) {
#endif
				if(!firstContact) cs << ",";
				cs << "(" << (seqLength-i) << "," << (seqLength-j) << ")";
				firstContact = false;
			}
		}
	}
	
	return cs.str();
}

void Conformation::addToContactFrequencyMatrix(vector< vector< int > > &freqMatrix, int count) {
	for(int i=0; i<seqLength; i++) {
		if( sequence[i] == P) continue;
				
		for(int j=i+2; j<seqLength; j++) {
			if( sequence[j] == P ) continue;
			
			int distance = abs(ii[i] - ii[j]);
			
			if( distance == 1 || distance == L ) {
				freqMatrix[i][j] += count;
				freqMatrix[j][i] += count;
			}
		}
	}
}

double Conformation::relativeHHContactOrder() {
	int sumOfDistance = 0;
	int n = 0;
	int l = 0;
	
	for(int i=0; i<seqLength; i++) {
		if( sequence[i] == P) continue;
		n++;
		
		for(int j=i+2; j<seqLength; j++) {
			if( sequence[j] == P ) continue;
			
			int distance = abs(ii[i] - ii[j]);

#if DIM == 2			
			if( distance == 1 || distance == L ) {
				l++;
				sumOfDistance += abs(i-j);
			}
#elif DIM == 3
			if( distance == 1 || distance == L || distance == LL ) {
				l++;
				sumOfDistance += abs(i-j);
			}
#endif
		}
	}
	if( n == 0 || l == 0) return 0.0;
	
	return (1.0 / (double)(l * n)) * sumOfDistance;
}

double Conformation::absoluteHHContactOrder() {
	int n = 0;
	for(int i=0; i<seqLength; i++) {
		if( sequence[i] != P ) n++;
	}
	
	return n * relativeHHContactOrder();
}

int Conformation::solventAccessibility() {
	int saCnt = 0;
	
	for(int i=0; i<seqLength; i++) {
		if( sequence[i] == P ) continue;
		
		for(int j=0; j<2*DIM; j++) {
			if( Lattice[ii[i] + DI[j]] == LPempty ) {
				saCnt++;
			}
		}
	}
	
	return saCnt;
}


void Conformation::loadConformation(const string &relDirections, const int freeEnergy) {
	int relDir[MAX_SEQ_LENGTH];
	int len = relDirections.size();
	for(int i=0; i<len; i++) {
		switch(relDirections[i]) {
			case 's': 	relDir[i] = straight; 	break;
			case 'l':	relDir[i] = left; 		break;
			case 'r':	relDir[i] = right; 		break;
			case 'u':	relDir[i] = up; 		break;
			case 'd':	relDir[i] = down; 		break;
			case 'g':	relDir[i] = ground; 	break;
		}
	}
	loadConformation(relDir, freeEnergy);
}

#if DIM == 3
void Conformation::loadConformation(const int relDirections[], const int freeEnergy) {
	reset(); // reset conformation
	//assert(testFreeEnergy(0));
	// set first two residues as the grounded residues
    Directions[0] = relDirections[0];
    Directions[1] = relDirections[1];
	Lattice[i0] = sequenceCode[0]; ii[0] = i0;
	Lattice[i0+1] = sequenceCode[1]; ii[1] = i0+1;
	
		   // s, r, u
	int basis[3] = {1, L, LL};
	int delta;
	
	for(int i=2; i<seqLength; i++) {
		Directions[i] = relDirections[i];
		
		switch(Directions[i]) {
			case straight:
				delta = basis[0]; 
				break;
			
			case right:
				delta = basis[1];
				basis[1] = -basis[0];
				break;
			
			case left:
				delta = -basis[1];
				basis[1] = basis[0];
				break;
			
			case up:
				delta = basis[2];
				basis[2] = -basis[0];
				break;
			
			case down:
				delta = -basis[2];
				basis[2] = basis[0];
				break;
		}
		basis[0] = delta;
		ii[i] = ii[i-1] + delta;
		Lattice[ii[i]] = sequenceCode[i];
	}
	computeFreeEnergy();
	//FreeEnergy = freeEnergy;
	searchResult.setMinimumSolution(Directions, FreeEnergy, seqLength);
	bestFreeEnergy = FreeEnergy;
	for(int i=0; i<seqLength; i++) {
		bestRelDirections[i] = Directions[i];
		ii_M[i] = ii[i];
		Lattice_M[ii_M[i]] = Lattice[ii[i]];
	}
}

#elif DIM == 2
void Conformation::loadConformation(const int relDirections[], const int freeEnergy) {
	reset(); // reset conformation
	//assert(testFreeEnergy(0));
	// set first two residues as the grounded residues
    Directions[0] = relDirections[0];
    Directions[1] = relDirections[1];
	Lattice[i0] = sequenceCode[0]; ii[0] = i0;
	Lattice[i0+L] = sequenceCode[1]; ii[1] = i0+L;
	
	int lp, dv;
	
	for(int i=2; i<seqLength; i++) {
		dv = ii[i-1] - ii[i-2];
		if( relDirections[i] == straight ) {
			lp = ii[i-1] + dv; 
		} else if( relDirections[i] == left ) {

			if( dv == L ) {
				lp = ii[i-1] + 1;
			} else if ( dv == -L ) {
				lp = ii[i-1] - 1;
			} else if ( dv == 1 ) {
				lp = ii[i-1] - L;
			} else if( dv == -1 ) {
				lp = ii[i-1] + L;
			} 					
		} else if( relDirections[i] == right ) {
			
			if( dv == L ) {
				lp = ii[i-1] - 1;
			} else if ( dv == -L ) {
				lp = ii[i-1] + 1;
			} else if ( dv == 1 ) {
				lp = ii[i-1] + L;
			} else if( dv == -1 ) {
				lp = ii[i-1] - L;
			} //else { assert(false); }
		} //else { assert(false); }
		Lattice[lp] = sequenceCode[i]; ii[i] = lp;
		Directions[i] = relDirections[i];
	}
	
	computeFreeEnergy();
	//FreeEnergy = freeEnergy;
	searchResult.setMinimumSolution(Directions, FreeEnergy, seqLength);
	bestFreeEnergy = FreeEnergy;
	for(int i=0; i<seqLength; i++) {
		bestRelDirections[i] = Directions[i];
		ii_M[i] = ii[i];
		Lattice_M[ii_M[i]] = Lattice[ii[i]];
	}
}

void Conformation::fastFold(double timeout) {
	cout << "Reseeding worst conformation - FreeEnergy: " << FreeEnergy << " -> ";
	makeStraight();		// reset the conformation
	stopwatch timer;
	timer.start();
	
	double oldTemp = getTemperature();
	setTemperature(190);
	
	while( timer.read() < timeout ) {
		runSimulation(1000);
	}

	loadConformation(bestRelDirections, bestFreeEnergy);
	setTemperature(oldTemp);
	cout << FreeEnergy << endl;
}

void Conformation:: printConformation() {
	// first we create a new lattice with additional coding
	// to make drawing easier
	// the residues are added to the lattice in sequence
	// odd numbers represent hydrophobic, even are polar.
	// ex.
	// For a sequence of the first five residues of the 
	// following sequence 'hhphp' would be:
	// 1,3,4,5.
	// explanation: residue #1 could be either a 1 or 2
	// depending on its HP property, residue #2 could be
	// 3 or 4, residue #n could be 2*n-1 or 2*n
	
	
	for(int i=0; i<HashSize; i++) {
		Lattice_M[i] = 0;
	}
	
	int code;
	for(int i=0; i<seqLength; i++) {
		if( sequence[i] == H ) {
			code = 2*(i+1) - 1;
		} else {
			code = 2*(i+1);
		}
		Lattice_M[ ii[i] ] = code;
	}
	
	// next we determine the region in the lattice 
	// that the fold resides so we don't need to draw the
	// entire lattice
	
	int col=0, minCol = L, maxCol = 0;
	int row=-1, minRow = L, maxRow = 0;
	for(int i=0; i<HashSize; i++) {
		if( i % L == 0 ) {
			col=0;
			row++;
		}
		if( Lattice[i] != LPempty ) {
			if( row < minRow ) minRow = row;
			if( row > maxRow ) maxRow = row;
			if( col < minCol ) minCol = col;
			if( col > maxCol ) maxCol = col;
		}
		col++;
	}
	
	cout << endl;
	
	for(int r=minRow; r<=maxRow; r++) {
		// first iteration of row to draw vertical bonds
		for(int c=minCol; c<=maxCol; c++) {
			if( r == 0 ) continue;		// don't process first row
			int lp = r * L + c;
			if( Lattice_M[lp] == 0 || Lattice_M[lp-L] == 0) {
				cout << "   ";
			} else {
				int curCode = Lattice_M[lp];
				int topCode = Lattice_M[lp-L];
				int diffCode = curCode - topCode;
				
				// is this a polar residue?
				if( (curCode % 2) == 0 ) {
					// is this the previous connected neighbor ?
					// or the next connected neighbour ?
					if( diffCode == 2 || diffCode == 3 || diffCode == -1 || diffCode == -2 ) {
						// draw a covalent vertical bond
						cout << "|  ";
					} else {
						cout << "   ";
					}
				} else {
					// is this the previous connected neighbor ?
					// or the next connected neighbour ?
					if( diffCode == 1 || diffCode == 2 || diffCode == -2 || diffCode == -3 ) {
						// draw a covalent vertical bond
						cout << "|  ";
					} else {
						// draw a vertical hydrogen bond if the top residue is hydrophobic
						if( SearchParameters::showHydrogenBonds && (topCode % 2) == 1 ) {
							cout << ":  ";
						} else {
							cout << "   ";
						}
					}
				}
			}
		}
		cout << endl;
			
		for(int c=minCol; c<maxCol; c++) {
			int lp = r * L + c;
			int curCode = Lattice_M[lp];
			
			if( curCode == 0 ) {
				cout << "   ";
				continue;
			}
			
			int rightCode = Lattice_M[lp+1];
			
			if( rightCode == 0 ) {
				if( (curCode % 2) == 0 ) {
					if( curCode == 2 || curCode == 2 * seqLength )
						cout << "P  ";
					else
						cout << "p  ";
				} else {
					if( curCode == 1 || curCode == 2 * seqLength - 1)
						cout << "H  ";
					else
						cout << "h  ";
				}
				continue;
			}
			
			int diffCode = curCode - rightCode;
			
			if( (curCode % 2) == 0 ) {
				if( curCode == 2 || curCode == 2 * seqLength )
					cout << "P";
				else
					cout << "p";
				
				// check for covalent bond
				if( diffCode == -1 || diffCode == -2 || diffCode == 2 || diffCode == 3 ) {
					cout << "--";
				} else {
					cout << "  ";
				}
			} else {
				if( curCode == 1 || curCode == 2 * seqLength - 1)
					cout << "H";
				else
					cout << "h";
				
				// check for covalent bond
				if( diffCode == 1 || diffCode == 2 || diffCode == -2 || diffCode == -3 ) {
					cout << "--";
				} else {
					// check for hydrogen bond
					if( SearchParameters::showHydrogenBonds && (rightCode % 2) == 1 ) {
						cout << "..";
					} else {
						cout << "  ";
					}
				}
			}
							
		}
			
		// now process last columns
		int lp = r * L + maxCol;
		int curCode = Lattice_M[lp];
			
		if( curCode == 0 ) {
			cout << " ";
		} else if( (curCode % 2) == 0 ) {
			if( curCode == 1 || curCode == 2 * seqLength - 1)
				cout << "P";
			else
				cout << "p";
		} else {
			if( curCode == 1 || curCode == 2 * seqLength - 1)
				cout << "H";
			else
				cout << "h";
		}
			
		cout << endl;
	}
	//for(int i=0; i<HashSize; i++) {
	//	if( i % L == 0 ) cout << endl;
	//	if( Lattice[i] == LPempty ) 
	//		cout << "?";
	///	else
	//		cout << Lattice[i];
	//}
	cout << endl;
}

#endif

void Conformation:: test2() {
/*	
	for(int i=0; i<seqLength; i++) {
		constrainedResidue[i] = true;
	}
	
	for(int i=0; i<11; i++) {
		constrainedResidue[i] = false;
	}
	
	for(int i=seqLength-12; i<seqLength; i++) {
		constrainedResidue[i] = false;
	}
	
	//printLattice();
	//cout << endl << "************************" << endl << endl;
	loadConformation("gglrrsrsllsslrrlslrrlslrrlrrlslrrlslrrlslrrlrrlrrlslrrlslrrlslslrrlslrrlssllsrsrrl",-31);
	runSimulation(10000000);
	loadConformation(bestRelDirections,bestFreeEnergy);
	//printLatticePositions();
	//printRelDirections();
	printConformation();
*/
/*
	printRelLatticePositions();
	printRelDirections();
	cout << "Lattice[ii[0]]=" << Lattice[ii[0]] << endl;
	
	tryPullMove(1);
	applyMutation( mutation );
	printRelLatticePositions();
	printRelDirections();
	assert(testLatticePositions());
	
	cout << "Lattice[ii[0]]=" << Lattice[ii[0]] << endl;
	
	tryPullMove(2);
	applyMutation( mutation );
	printRelLatticePositions();
	printRelDirections();
	assert(testLatticePositions());
	
	cout << "Lattice[ii[0]]=" << Lattice[ii[0]] << endl;
	
	tryPullMove(3);
	applyMutation( mutation );
	printRelLatticePositions();
	printRelDirections();
	assert(testLatticePositions());
	
	cout << "Lattice[ii[0]]=" << Lattice[ii[0]] << endl;
	
	tryPullMove(4);
	applyMutation( mutation );
	printRelLatticePositions();
	printRelDirections();
	assert(testLatticePositions());
	
	cout << "Lattice[ii[0]]=" << Lattice[ii[0]] << endl;
	
	tryPullMove(1);
	applyMutation( mutation );
	printRelLatticePositions();
	printRelDirections();
	assert(testLatticePositions());
	
	cout << "Lattice[ii[0]]=" << Lattice[ii[0]] << endl;
	
	tryPullMove(2);
	applyMutation( mutation );
	printRelLatticePositions();
	printRelDirections();
	assert(testLatticePositions());
	
	cout << "Lattice[ii[0]]=" << Lattice[ii[0]] << endl;
	
	tryPullMove(3);
	applyMutation( mutation );
	printRelLatticePositions();
	printRelDirections();
	assert(testLatticePositions());
	
	cout << "Lattice[ii[0]]=" << Lattice[ii[0]] << endl;
	
	tryPullMove(4);
	applyMutation( mutation );
	printRelLatticePositions();
	printRelDirections();
	assert(testLatticePositions());
	
	cout << "Lattice[ii[0]]=" << Lattice[ii[0]] << endl;
	
	tryPullMove(1);
	applyMutation( mutation );
	printRelLatticePositions();
	printRelDirections();
	assert(testLatticePositions());
	
	cout << "Lattice[ii[0]]=" << Lattice[ii[0]] << endl;
	
	tryPullMove(2);
	applyMutation( mutation );
	printRelLatticePositions();
	printRelDirections();
	assert(testLatticePositions());
	
	tryPullMove(3);
	applyMutation( mutation );
	printRelLatticePositions();
	printRelDirections();
	assert(testLatticePositions());
*/
}

