/*
 *  Project: Replica exchange Monte Carlo for Protein Folding in the HP model
 *  File:    Conformation.h
 *
 *  Note: The latest source code and information regarding this project 
 *  is available at http://www.cs.ubc.ca/labs/beta/Projects/REMC-HPPFP/
 *
 *  For citation purposes, please use:
 *  Thachuk C, Shmygelska A and Hoos H: Replica exchange Monte Carlo for 
 *  protein folding in the HP model. BMC Bioinformatics 2007, Submitted.
 * 
 *  Copyright (C) 2007  Chris Thachuk, Alena Shmygelska and Holger H Hoos
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */
 
#ifndef CONFORMATION_H
#define CONFORMATION_H

#include <iostream>
#include <vector>
#include <string>
#include <math.h>
#include <cstdlib>
#include <ctime>
#include <utility> 
#include <map>
#include <assert.h>
#include "Const.h"
#include "Random.h"
#include "SearchResult.h"
#include "SearchParameters.h"
#include "Timer.h"
#include "stopwatch.h"

// default static initialization
#if DIM == 2
static const int VSF[2] = {1,0};
static const int VLF[2] = {0,1};
static const int VRF[2] = {0,-1};
static const int VBF[2] = {-1,0};

static const int VSB[2] = {-1,0};
static const int VLB[2] = {0,-1};                                                                               
static const int VRB[2] = {0,1};
static const int VBB[2] = {1,0};

static const int BS[2][2] = {1,0, 0,1};
static const int BL[2][2] = {0,-1, 1,0};
static const int BR[2][2] = {0,1, -1,0};
static const int BB[2][2] = {-1,0,0,-1};

static const int DI[4] = {1,L,-1,-L};

#elif DIM == 3
static const int VSF[3] = {1,0,0};
static const int VLF[3] = {0,1,0};
static const int VRF[3] = {0,-1,0};
static const int VBF[3] = {-1,0,0};
static const int VUF[3] = {0,0,1};
static const int VDF[3] = {0,0,-1};

static const int VSB[3] = {-1,0,0};
static const int VLB[3] = {0,-1,0};                                                                               
static const int VRB[3] = {0,1,0};
static const int VBB[3] = {1,0,0};
static const int VUB[3] = {0,0,-1};
static const int VDB[3] = {0,0,1};

static const int BS[3][3] = {1,0,0,   0,1,0,  0,0,1};
static const int BL[3][3] = {0,-1,0, 1,0,0,   0,0,1};
static const int BR[3][3] = {0,1,0,  -1,0,0,  0,0,1};
static const int BU[3][3] = {0,0,-1, 0,1,0, 1,0,0};
static const int BD[3][3] = {0,0,1, 0,1,0, -1,0,0};
static const int BB[3][3] = {-1,0,0, 0,-1,0, 0,0,-1};

static const int DI[6] = {1,L,LL,-1,-L,-LL};

#endif

using namespace std;

class Conformation {
	private:
		
		static int OFFSET;
		static double probAcceptValues[MAX_NUM_REPLICAS][MAX_NUM_ENERGY_VALUES];
		static double Tvalues[MAX_NUM_REPLICAS];
		
		// local class properties
		int replicaID;
		int seqLength;
		int FreeEnergy;
		double pmWeight;
		double T;
		int Lattice[HashSize]; //Lattice hash
		int i0;	// initial starting position in lattice
		int ii[MAX_SEQ_LENGTH]; //hash keys into the lattice              
		int FixedPoint; //folding point
		int Directions[MAX_SEQ_LENGTH]; //relative directions of aa
						
		int sequence[MAX_SEQ_LENGTH];
		int sequenceCode[MAX_SEQ_LENGTH];
		int bestRelDirections[MAX_SEQ_LENGTH];
		int bestFreeEnergy;
		
		int FreeEnergy_M;
		int Lattice_M[HashSize];
		int ii_M[MAX_SEQ_LENGTH];
		int move_startResidue;
		int move_endResidue;
		bool move_isPossible;
		
#if MOVESET == PULLMOVES || MOVESET == MIXED
		// this structure is only for pullmoves
		// in this structure, values are stored as residue, lattice_position
		//map<int,int> newResidueMapping;
		
		// again this is only for pullmoves
		// we record the new residue mappings for each access in O(1) time
		// in this hashmap
		// values are always stored as lattice_position, residue,
		// not the other way around
		// entries will be blanked after each pull move
		//map<int,int> newResidueMappingR;
		
		// always we record the old mapping, so we can determine where a bond
		// should not be
		//map<int,int> oldResidueMappingR;
#endif

		//Mutation mutation;
		SearchResult searchResult;
		float prevTime;
		long double prevMCStep;
		Timer t_clock;
		
		// used to track statistics
		long double acceptedMoves;
		long double rejectedMoves;
		long double impossibleMoves;
		long double numResidues;
		
		// private class methods
		int row(int);
		int col(int);
#if DIM == 3
		int height(int);
#endif
		void determineDirection(int i, int n1, int n2, int &Dir, int GBasisP[DIM][DIM], int fixedpoint);
#if DIM == 3
		int determineII(int TBasis [DIM][DIM], int Direction, bool forward,int newBasis[DIM][DIM]);
#endif
		void determineBasis( int n1, int n2, int Basis[DIM][DIM] );
		int findFreeNeighbours(int residue, int freeN[]);
		int findFreeNeighboursL(int lp, int freeN[DIM*2] );
		int findFreeNeighbours(int res1, int res2, int freeN[DIM*2]);
		int selectRandomFreeNeighbour(int freeN[DIM*2], int numFree);
		int countBonds(int n1, int lp, int plp, int nlp);
		void applyMutation();
		void resetMutation();
		void recenterOnLattice();

#if MOVESET == LOCAL || MOVESET == MIXED
		void tryEndMove(int residue);
//#if DIM == 2
//		void tryCrankshaftOrCornerMove(int residue);
//#elif DIM == 3
		void tryCrankshaftMove(int residue);
		void tryCornerMove(int residue);
//#endif
#endif


#if MOVESET == PULLMOVES || MOVESET == MIXED
		void tryEndPullMove(int residue);
		void tryPullMove(int residue);
#endif
		
		// private methods
		void printRelDirections();
		void printLatticePositions();
		
		int calcFreeEnergy();
		
	// public interface
	public:
		// default constructor
		Conformation(const string &sequence, const double temperature, const string &experimentID, const int replicaNumber);
		
		static void initProbAcceptValues(const double tempValues[], int numReplicas, int absKnownEnergy) {
			// for each temperature value
			for(int i=0; i<numReplicas; i++) {
				double T = tempValues[i];
				Tvalues[i] = T;
				
				// for each possible energy delta
				for(int j=0; j<=absKnownEnergy; j++) {
					probAcceptValues[i][j] = exp(-j/(T*K_b));
				}
			}
			
		}
		
		static void swapOffset() {
			OFFSET = 1 - OFFSET;
		}
		
		bool testLatticePositions();
		bool testFreeEnergy(int);
		void printRelLatticePositions();
		
		// method to set the sequence
		void setSequence( const string &seq ) {
			seqLength = seq.length();
			for(int i=0; i<seqLength; i++) {
				if( seq[i] == 'h')
					sequence[i] = H;
				else
					sequence[i] = P;
			}
		}
		
		// method to get the sequence
		string getSequence() {
			ostringstream buf;
			for(int i=0; i<seqLength; i++) {
				if( sequence[i] == H )
					buf << 'h';
				else
					buf << 'p';
			}
			return buf.str();
		}

		string getDirections() {
			char relDir[seqLength];
		  	for(int i=0; i<seqLength; i++) {
				switch(Directions[i]) {
				case straight: relDir[i] = 's'; 	break;
				case left:	 relDir[i] = 'l'; 		break;
				case right:	 relDir[i] = 'r'; 		break;
				case up:	 relDir[i] = 'u'; 		break;
				case down:	 relDir[i] = 'd'; 		break;
				case ground: relDir[i] = 'g'; 	break;
				}
		  	}
			string tmp = relDir;
			return tmp;
		}
		
		void recalculateDirections();
		
		// returns the current temperature of the system
		double getTemperature();
		
		// set the current temperature of the system
		void setTemperature(const double t);
		
		// returns the current free energy of the system
		int getFreeEnergy();
		
		// returns the replica id
		int getReplicaID();

		void setReplicaID(int a) {
			replicaID = a;
		}
		
		// runs the simulation for n steps
		SearchResult& runSimulation(const long double n);
		
		// returns the search results
		SearchResult& getSearchResult();
		
		// informs SearchResult to dump results to file
		void writeResults();
		
		// temporary test function
		void test2();
		
		// blank the lattice and all conformation information
		void reset();
		
		// reset the conformation to a straight conformation
		void makeStraight();
		
		// calculate the free energy of current conformation
		void computeFreeEnergy();
		
		// prints the conformation
		void printConformation();
		
		// returns the contact string
		string getContactString();
		
		// returns the symmetric contact string in case folding occured from the other end
		// in a symmetric sequence
		string getSymmetricContactString();
		
		// returns the relative h-h contact order 
		double relativeHHContactOrder();
		
		// returns the absolute h-h contact order
		double absoluteHHContactOrder();
		
		// returns the solvent accessible area
		int solventAccessibility();
		
		// adds frequency of residue contacts to contact frequency matrix (parameter)
		void addToContactFrequencyMatrix(vector< vector< int > > &, int);
		
#if DIM == 2		
		// quickly folds a near optimal solution
		void fastFold(double);
#endif		
		// loads a conformation given by the parameters
		void loadConformation(const int [], const int=0);
		
		void loadConformation(const string &, const int=0);
		

};

// returns the relative row of a lattice position
inline
int Conformation:: row(int latticePos) {
#if DIM == 2
	return latticePos / L;
#elif DIM == 3
	return (latticePos & (LL-1)) / L; 
#endif
}

// returns the relative column of a lattice position
inline
int Conformation:: col(int latticePos) {
	return latticePos & (L - 1);
}

#if DIM == 3
// returns the relative height of a lattice position
inline
int Conformation:: height(int latticePos) {
	return latticePos / LL;
}
#endif

#if DIM == 2
// given two residue n1 and n2, determines the basis of direction
// from n1 to n2 in 2D
inline
void Conformation:: determineBasis( int n1, int n2, int Basis[DIM][DIM] ) {
	int vect[DIM];
	const int *basis;
	
	if( n2>n1 ) {
		vect[0] = row(ii[n2]) - row(ii[n1]);
		vect[1] = col(ii[n2]) - col(ii[n1]);
		
		if( vect[0] == VSF[0]  && vect[1] == VSF[1] ) {
			basis = &BS[0][0];
		} else if( vect[0] == VLF[0]  && vect[1] == VLF[1] ) {
			basis = &BL[0][0];
		} else if( vect[0] == VRF[0]  && vect[1] == VRF[1] ) {
			basis = &BR[0][0];
		} else if( vect[0] == VBF[0]  && vect[1] == VBF[1] ) {
			basis = &BB[0][0];
		}
	} else {
		vect[0] = row(ii[n1]) - row(ii[n2]);
		vect[1] = col(ii[n1]) - col(ii[n2]);
		
		if( vect[0] == VSB[0]  && vect[1] == VSB[1] ) {
			basis = &BS[0][0];
		} else if( vect[0] == VLB[0]  && vect[1] == VLB[1] ) {
			basis = &BL[0][0];
		} else if( vect[0] == VRB[0]  && vect[1] == VRB[1] ) {
			basis = &BR[0][0];
		} else if( vect[0] == VBB[0]  && vect[1] == VBB[1] ) {
			basis = &BB[0][0];
		}
		
	}
	
	for(int row=0; row<DIM; row++) {
		for(int col=0; col<DIM; col++) {
			Basis[row][col] = *(basis + row*DIM + col);
			//cout << Basis[row][col] << " ";
		}
		//cout << endl;
	}
}

#elif DIM == 3

// given two residue n1 and n2, determines the basis of direction
// from n1 to n2 in 3D
inline
void Conformation:: determineBasis( int n1, int n2, int Basis[DIM][DIM] ) {
	int vect[DIM];
	const int *basis;
	
	if( n2>n1 ) {
		vect[0] = height(ii[n2]) - height(ii[n1]);
		vect[1] = row(ii[n2]) - row(ii[n1]);
		vect[2] = col(ii[n2]) - col(ii[n1]);
		
		if( vect[0] == VSF[0]  && vect[1] == VSF[1] && vect[2] == VSF[2] ) {
			basis = &BS[0][0];
		} else if( vect[0] == VLF[0]  && vect[1] == VLF[1] && vect[2] == VLF[2] ) {
			basis = &BL[0][0];
		} else if( vect[0] == VRF[0]  && vect[1] == VRF[1] && vect[2] == VRF[2] ) {
			basis = &BR[0][0];
		} else if( vect[0] == VUF[0]  && vect[1] == VUF[1] && vect[2] == VUF[2] ) {
			basis = &BU[0][0];
		} else if( vect[0] == VDF[0]  && vect[1] == VDF[1] && vect[2] == VDF[2] ) {
			basis = &BD[0][0];
		} else if( vect[0] == VBF[0]  && vect[1] == VBF[1] && vect[2] == VBF[2] ) {
			basis = &BB[0][0];
		} 
	} else {
		vect[0] = height(ii[n1]) - height(ii[n2]);
		vect[1] = row(ii[n1]) - row(ii[n2]);
		vect[2] = col(ii[n1]) - col(ii[n2]);
		
		if( vect[0] == VSB[0]  && vect[1] == VSB[1] && vect[2] == VSB[2] ) {
			basis = &BS[0][0];
		} else if( vect[0] == VLB[0]  && vect[1] == VLB[1] && vect[2] == VLB[2] ) {
			basis = &BL[0][0];
		} else if( vect[0] == VRB[0]  && vect[1] == VRB[1] && vect[2] == VRB[2] ) {
			basis = &BR[0][0];
		} else if( vect[0] == VUB[0]  && vect[1] == VUB[1] && vect[2] == VUB[2] ) {
			basis = &BU[0][0];
		} else if( vect[0] == VDB[0]  && vect[1] == VDB[1] && vect[2] == VDB[2] ) {
			basis = &BD[0][0];
		} else if( vect[0] == VBB[0]  && vect[1] == VBB[1] && vect[2] == VBB[2] ) {
			basis = &BB[0][0];
		}
		
	}
	
	for(int row=0; row<DIM; row++) {
		for(int col=0; col<DIM; col++) {
			Basis[row][col] = *(basis + row*DIM + col);
			//cout << Basis[row][col] << " ";
		}
		//cout << endl;
	}
}

//---------------------------------------------------------------------------------------------
//determine coordinates given previous basis and current direction, moving back/forward
inline
int Conformation::determineII(int TBasis [DIM][DIM], int Direction, bool forward,int newBasis[DIM][DIM]){  
	int ii= 0;
	int vect[DIM];
	int dispVect[DIM];
	int multBasis[DIM][DIM];
	int prevBasis[DIM][DIM];
	
	switch(Direction){
		
	    case straight:
			for(int row = 0; row<DIM; row++){
				if(forward)
					vect[row] = VSF[row];
				else
					vect[row] = VSB[row];
				for(int col = 0; col<DIM; col++)
					multBasis[row][col] = BS[row][col];
				
			}	      
			
			break;
			
	    case left:
			for(int row = 0; row<DIM; row++){
				if(forward)
					vect[row] = VLF[row];
				else
					vect[row] = VLB[row];
				for(int col = 0; col<DIM; col++)
					multBasis[row][col] = BL[row][col];
				
			}	      
			
			break;
			
	    case right:
			for(int row = 0; row<DIM; row++){
				if(forward)
					vect[row] = VRF[row];
				else
					vect[row] = VRB[row];
				for(int col = 0; col<DIM; col++)
					multBasis[row][col] = BR[row][col];
				
			}	      
			
			break;

	    case up:
			for(int row = 0; row<DIM; row++){
				if(forward)
					vect[row] = VUF[row];
				else
					vect[row] = VUB[row];
				for(int col = 0; col<DIM; col++)
					multBasis[row][col] = BU[row][col];
				
			}	      
			
			break;
			
	    case down:
			
			for(int row = 0; row<DIM; row++){
				if(forward)
					vect[row] = VDF[row];
				else
					vect[row] = VDB[row];
				for(int col = 0; col<DIM; col++){
					multBasis[row][col] = BD[row][col];
					
				}
				
			}	      
			
			break;
		
		default:
			break;
	}
	
	   int Temp[DIM][DIM];
	for(int row=0; row<DIM; row++){
		for(int col=0; col<DIM; col++){
			Temp[row][col] = 0;
			for(int mf=0; mf<DIM; mf++)
				Temp[row][col]+=TBasis[row][mf]*multBasis[mf][col];
			
		}
		
	}
	
	
	for(int row = 0; row<DIM; row++){
		for(int col = 0; col<DIM; col++)
	       {
			prevBasis[row][col] = TBasis[row][col];
			newBasis[row][col] = Temp[row][col];             
			
		   }
		
	}
	
	for(int row = 0; row<DIM; row++){
		dispVect[row] = 0;
		for(int col = 0; col<DIM; col++){
			dispVect[row]+=prevBasis[row][col]*vect[col];
		}
		
		
	}
	
	ii =dispVect[0]*LL+dispVect[1]*L+dispVect[2];

	return ii;
}            


#endif

// given a free neighbour matrix, randomly selects one of the free positions
inline
int Conformation:: selectRandomFreeNeighbour(int freeN[DIM*2], int numFree) {
	double cum_p[DIM*2];
	int Q = random_number(1, numFree);
	
	int index = 0;
	int passes = 0;
	
	for(int i=0; i<DIM*2; i++) {
		if( freeN[i] != -1 ) {
			passes++;
		}
		if( passes == Q ) {
			index = i;
			break;
		}
	}
	
	return freeN[index];
}

// returns the free neighbours of a given residue
inline
int Conformation:: findFreeNeighbours(int residue, int freeN[DIM*2]) {
	return findFreeNeighboursL(ii[residue], freeN);
}

inline
int Conformation:: findFreeNeighboursL(int lp, int freeN[DIM*2] ) {
	int numFree = 0;
	
	for(int i=0; i<DIM*2; i++) {
		if( Lattice[lp + DI[i]] == LPempty ) {
			freeN[i] = lp + DI[i];
			numFree++;
		} else {
			freeN[i] = -1;
		}
	}
	return numFree;
}

// returns the free neighbours in common of two given residues
inline
int Conformation:: findFreeNeighbours(int res1, int res2, int freeN[DIM*2]) {
	int freeN1[DIM*2];
	int freeN2[DIM*2];
	int numFree=0;
	
	findFreeNeighbours(res1, freeN1);
	findFreeNeighbours(res2, freeN2);
	
	for(int i=0; i<DIM*2; i++) {
		freeN[i]=-1;
		if( freeN1[i] != -1 ) {
			for(int j=0; j<DIM*2; j++) {
				if( freeN1[i] == freeN2[j] ) {
					freeN[i] = freeN1[i];
					numFree++;
				}
			}
		}
	}
	return numFree;
}

// returns the number of hydrogen bonds between residue at 
// lattice position lp and other residues.  plp and nlp are 
// the lattice positions of the previous and next residue.
// these are required to ensure we don't count them as hydrogen
// bonds.  if plp or nlp = -1, we assume they don't exist
inline
int Conformation::countBonds(int n1, int lp, int plp, int nlp) {
	int bonds = 0;
	
	if( sequence[n1] < 0 ) return 0;
	
	for(int i=0; i < 2*DIM; i++) {
		int olp = lp + DI[i];
		if( Lattice[ olp ] >= 0 && olp != plp && olp != nlp && olp != ii[n1] ) bonds++; 
	}
	return bonds;
}

#if DIM == 2
// returns the new direction from n1 to n2 in 2D
inline
void Conformation:: determineDirection(int i, int n1, int n2, 
	int &Dir, int GBasisP[DIM][DIM], int fixedpoint) {
	
 	int rvect[DIM];
	int co_i[DIM], co_j[DIM];
	
	if(i==fixedpoint-1 || i == fixedpoint-2){
		Dir=ground;
		GBasisP[0][0] = 1;
		GBasisP[0][1] = 0;
		
		GBasisP[1][0] = 0;
		GBasisP[1][1] = 1;
		
	}else {
		if(i==n2){
			
			co_i[0]=ii[i]/L;
			co_i[1]=(ii[i]&(L-1));
			co_j[0]=ii[i-1]/L;
			co_j[1]=(ii[i-1]&(L-1));
			
			for(int dim = 0; dim<DIM; dim++)
				rvect[dim]= co_i[dim]-co_j[dim];
			
			int vect[DIM]; 
            int multBasis [DIM][DIM];
			
            for (int row = 0; row<DIM; row++){
				int transp_row[DIM];
				
				for(int col = 0; col<DIM; col++){
					transp_row[col] = GBasisP[col][row];
				}
				
				vect[row] = 0;
				for(int col = 0; col<DIM; col++){
					vect[row]+=transp_row[col]*rvect[col];
				}
			}
            
			if(((vect[0] == VSF[0]) && (vect[1] == VSF[1]))){// && (vect[2] == VSF[2])) ){
				Dir = straight;
				
				for(int row = 0; row<DIM; row++){
					for(int col = 0; col<DIM; col++){
						multBasis[row][col] = BS[row][col];
					}
                }
			}
			else if((vect[0] == VLF[0] && vect[1] == VLF[1] )){//&& vect[2] == VLF[2])  ){
                Dir = left;
				
				for(int row = 0; row<DIM; row++)
					for(int col = 0; col<DIM; col++)
						multBasis[row][col] = BL[row][col];
			}
			else if((vect[0] == VRF[0] && vect[1] == VRF[1])){// && vect[2] == VRF[2])){
				Dir = right;
				for(int row = 0; row<DIM; row++)
					for(int col = 0; col<DIM; col++)
						multBasis[row][col] = BR[row][col];
			}
			
			int prevBasis[DIM][DIM];
            for(int row = 0; row<DIM; row++)
				for(int col = 0; col<DIM; col++)
					prevBasis[row][col] = GBasisP[row][col];
			for(int row=0; row<DIM; row++){
				for(int col=0; col<DIM; col++){
					GBasisP[row][col] = 0;
					for(int mf=0; mf<DIM; mf++)
						GBasisP[row][col]+=prevBasis[row][mf]*multBasis[mf][col];
				}
			}
		}
		
	} 
}

#elif DIM == 3
// returns the new direction from n1 to n2 in 3D
inline
void Conformation:: determineDirection(int i, int n1, int n2, 
	int &Dir, int GBasisP[DIM][DIM], int fixedpoint){
	
 	int rvect[DIM];
	int co_i[DIM], co_j[DIM];
	
	if(i==fixedpoint-1 || i == fixedpoint-2){
		Dir=ground;
		GBasisP[0][0] = 1;
		GBasisP[0][1] = 0;
		GBasisP[0][2] = 0;
		
		GBasisP[1][0] = 0;
		GBasisP[1][1] = 1;
		GBasisP[1][2] = 0;
		
		GBasisP[2][0] = 0;
		GBasisP[2][1] = 0;
		GBasisP[2][2] = 1;
		
	}else {
		if(i==n2){
			
			co_i[0]=ii[i]/LL;
			co_i[1]=(ii[i]&(LL-1))/L;
			co_i[2] = ii[i]&(L-1);
			co_j[0]=ii[i-1]/LL;
			co_j[1]=(ii[i-1]&(LL-1))/L;
			co_j[2] = ii[i-1]&(L-1);
			
			for(int dim = 0; dim<DIM; dim++)
				rvect[dim]= co_i[dim]-co_j[dim];
			
			int vect[DIM]; 
            int multBasis [DIM][DIM];
			
            for (int row = 0; row<DIM; row++){
				int transp_row[DIM];
				
				for(int col = 0; col<DIM; col++){
					transp_row[col] = GBasisP[col][row];
				}
				
				vect[row] = 0;
				for(int col = 0; col<DIM; col++){
					vect[row]+=rvect[col]*transp_row[col];
					
				}
			}
            
			if( (vect[0] == VSF[0]) && (vect[1] == VSF[1]) && (vect[2] == VSF[2]) ) {
				Dir = straight;
				
				for(int row = 0; row<DIM; row++)
					for(int col = 0; col<DIM; col++)
						multBasis[row][col] = BS[row][col];
			}
			else if( (vect[0] == VLF[0]) && (vect[1] == VLF[1]) && (vect[2] == VLF[2]) ){
                Dir = left;
				
				for(int row = 0; row<DIM; row++)
					for(int col = 0; col<DIM; col++)
						multBasis[row][col] = BL[row][col];
			}
			else if( (vect[0] == VRF[0]) && (vect[1] == VRF[1]) && (vect[2] == VRF[2]) ) {
				Dir = right;
				for(int row = 0; row<DIM; row++)
					for(int col = 0; col<DIM; col++)
						multBasis[row][col] = BR[row][col];
			}
			else if( (vect[0] == VUF[0]) && (vect[1] == VUF[1]) && (vect[2] == VUF[2]) ) {
				Dir = up;
				for(int row = 0; row<DIM; row++)
					for(int col = 0; col<DIM; col++)
						multBasis[row][col] = BU[row][col];
			}
			else if( (vect[0] == VDF[0]) && (vect[1] == VDF[1]) && (vect[2] == VDF[2]) ) {
				Dir = down;
				for(int row = 0; row<DIM; row++)
					for(int col = 0; col<DIM; col++)
						multBasis[row][col] = BD[row][col];
			}
			
			int prevBasis[DIM][DIM];
            for(int row = 0; row<DIM; row++)
				for(int col = 0; col<DIM; col++)
					prevBasis[row][col] = GBasisP[row][col];
					
			for(int row=0; row<DIM; row++){
				for(int col=0; col<DIM; col++){
					GBasisP[row][col] = 0;
					for(int mf=0; mf<DIM; mf++)
						GBasisP[row][col]+=prevBasis[row][mf]*multBasis[mf][col];
				}
			}
		}
		
	} 
}
#endif

#endif
