/*
 *  Project: Replica exchange Monte Carlo for Protein Folding in the HP model
 *  File:    REMSimulation.cpp
 *
 *  Note: The latest source code and information regarding this project 
 *  is available at http://www.cs.ubc.ca/labs/beta/Projects/REMC-HPPFP/
 *
 *  For citation purposes, please use:
 *  Thachuk C, Shmygelska A and Hoos H: Replica exchange Monte Carlo for 
 *  protein folding in the HP model. BMC Bioinformatics 2007, Submitted.
 * 
 *  Copyright (C) 2007  Chris Thachuk, Alena Shmygelska and Holger H Hoos
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include "REMSimulation.h"

REMSimulation::REMSimulation( const SearchParameters &params ) {
	this->params = params;
	
	// if explicit temperature values were not given
	if( params.temperatureValues.size() == 0 ) {
		// start first replica temp at the minimum
		double currentTemp = params.minTemperature;
		// evenly distribute the temperature increment amongst the replicas
		double tempStep = params.numReplicas>1?(params.maxTemperature - params.minTemperature) / (params.numReplicas - 1):1;
		numReplicas = params.numReplicas;
		
		// create linear temperature map
		for( int i = 0; i < params.numReplicas; i++ ) {
			tempValues[i] = currentTemp;
			currentTemp += tempStep;
		}
	} 
	// if explicit temperature values were given
	else {
		vector<double>::iterator itr;
		int i = 0;
		params.numReplicas = params.temperatureValues.size();
		numReplicas = params.numReplicas;
		
		for(itr = params.temperatureValues.begin(); itr != params.temperatureValues.end(); itr++) {
			tempValues[i] = *itr;
			i++;
		}
	}
	
	// create all replicas in increasing temperature order
	for( int i = 0; i < params.numReplicas; i++ ) {
		Replica *r;
		r = new Replica( params.sequence, tempValues[i], params.experimentID, i );
		replicas.push_back( r );
		
		if( params.initConf != "" ) {
			cout << "Loading initial Conformation with Free Energy: ";
			r->loadConformation(params.initConf);
			r->computeFreeEnergy();
			cout << r->getFreeEnergy() << endl;
		}

#if defined TRACE_REP_SWAPS || defined TRACE_ARI_MOVES
		tempMap[tempValues[i]] = i;
#endif	
	}
	
	// OPTIMIZATION -- store pre-computed delta values for fast lookup
	for(int i=0; i<params.numReplicas - 1; i++) {
		double B1 = 1.0 / (K_b2 * tempValues[i]);
		double B2 = 1.0 / (K_b2 * tempValues[i+1]);
		double B_diff = B2 - B1;
		
		// generate delta value for each possible energy delta
		for(int j=0; j<=abs(params.knownEnergy); j++) {
			double delta = B_diff * (-j);
			probDeltaValues[i][j] = exp(-delta);
		}
	}
	
	// OPTIMIZATION -- store pre-computed acceptance probabilities for fast lookup
	// in Conformation class
	Conformation::initProbAcceptValues(tempValues, params.numReplicas, abs(params.knownEnergy));
	
}

REMSimulation::~REMSimulation() {

	for(int i = 0; i < params.numReplicas; i++ ) {
		delete replicas[i];
	}
}

bool REMSimulation::shouldSwapReplicas( int i, int j ) {
	
	int Ei, Ej, E_delta;
	
	Ei = replicas[i]->getFreeEnergy();
	Ej = replicas[j]->getFreeEnergy();

	E_delta = Ei - Ej;
	
	if( E_delta > 0 ) {
		return true;
	} else {
		double z = ran01();
		
		if( probDeltaValues[i][-E_delta] > z )
			return true;
		else
			return false;
	}
}

void REMSimulation::swapReplicas( int i, int j ) {
	Conformation *tmp = replicas[i];
	replicas[i] = replicas[j];
	replicas[j] = tmp;
	
	replicas[i]->setReplicaID(i);
	replicas[j]->setReplicaID(j);	
}

void REMSimulation::runSimulation(vector<SQT_item> &bestSolutionTrace) {
	int prevBestSolution, bestSolution = 0;
	int exch = 0;
	bool cool = true;
	string bestRelDirections;

#ifdef RESEEDING
	int sinceReseed = 0;
#endif
	
#ifdef TRACE_SQT_GENERAL_IMMEDIATE
	ostringstream sqt_filename;
	sqt_filename << params.traceFile;
	ofstream sqt;
	sqt.open(sqt_filename.str().c_str(),ios::out);
#endif
	
#ifdef TRACE_REP_SWAPS
	int repSwaps[params.numReplicas][params.numReplicas][2];
	for(int i=0; i<params.numReplicas; i++) {
		for(int j=0; j<params.numReplicas; j++) {
			for(int k=0; k<2; k++) {
				repSwaps[i][j][k]=0;
			}
		}
	}
#endif

#ifdef TRACE_ARI_MOVES
	long double ariMoves[params.numReplicas][5];	// [i][0] -> acceptedMoves
													// [i][1] -> rejectedMoves
													// [i][2] -> impossibleMoves
													// [i][3] -> totalMoves
													// [i][4] -> total number of residues involved
													// for all i
	for(int i=0; i<params.numReplicas; i++) {
		for(int j=0; j<4; j++) {
			ariMoves[i][j] = 0;
		}
	}
#endif
	
	//Timer timer;
	//timer.start();
	stopwatch timer;
	timer.start();
		
	while( bestSolution > params.knownEnergy && timer.read() < params.maxRunTime ) {
		
		ReplicaIterator replicaItr;
		int repId = 0;
		int bestRep = 0;
		int bestEnergy = 0;
		int worstRep = 0;
		int worstEnergy = -99;		

		// run each replica in simulation for fixed number of steps
		for( replicaItr = replicas.begin(); replicaItr != replicas.end(); replicaItr++ ) {
			Replica *replica = *replicaItr;
#if defined ANNEALING
			if( cool ) {
				replica->setTemperature( replica->getTemperature() / SearchParameters::mu );
			} else {
				replica->setTemperature( replica->getTemperature() * SearchParameters::mu );
			}
#endif
			SearchResult &sr = replica->runSimulation( params.numLocalSteps );
			SQT_item &ms_sqt = sr.getMinimumSolution();
			
			if( replica->getFreeEnergy() < bestEnergy ) {
				bestEnergy = replica->getFreeEnergy();
				bestRep = repId;
			}
			
			if( replica->getFreeEnergy() > worstEnergy ) {
				worstEnergy = replica->getFreeEnergy();
				worstRep = repId;
			}
			
			if( ms_sqt.freeEnergy < bestSolution ) {
				bestSolution = ms_sqt.freeEnergy;
				ms_sqt.runTime = timer.read();
				bestRelDirections = ms_sqt.relDirections;
#if defined TRACE_SQT_GENERAL_CACHED
				SQT_item sqt_item = ms_sqt;
				//sqt_item.relDirections = "";
				bestSolutionTrace.push_back(ms_sqt);

#elif defined TRACE_SQT_GENERAL_IMMEDIATE
				sqt << ms_sqt.runTime << "\t" << bestSolution << "\t" << ms_sqt.relDirections << endl;
#endif
			}

#ifdef TRACE_ARI_MOVES
			ariMoves[repId][0] += sr.getAcceptedMoves();
			ariMoves[repId][1] += sr.getRejectedMoves();
			ariMoves[repId][2] += sr.getImpossibleMoves();
			ariMoves[repId][3] += sr.getNumSteps();
			ariMoves[repId][4] += sr.getNumResidues();
#endif
			repId++;
		}
		
#if defined ANNEALING
		if( cool ) {
			if( replicas[params.numReplicas-1]->getTemperature() < params.minTemperature ) {
				cool = false;
				cerr << "Global Heating (" << replicas[0]->getFreeEnergy() << "," << replicas[params.numReplicas-1]->getFreeEnergy() << ")" << endl;
			}
		} else {
			if( replicas[0]->getTemperature() > params.maxTemperature ) {
				cool = true;
				cerr << "Global Cooling (" << replicas[0]->getFreeEnergy() << "," << replicas[params.numReplicas-1]->getFreeEnergy() << ")" << endl;
			}
		}
#endif
		
#if defined RESEEDING
		sinceReseed++;
		if(sinceReseed > params.reseedAfter) {
			replicas[worstRep]->fastFold(params.reseedTimeout);
			//replicas[bestRep]->fastFold(params.reseedTimeout);
			//for(int k=0; k<params.numReplicas; k++) {
			//	if(k==bestRep) continue;
			//	replicas[k]->fastFold(params.reseedTimeout);
			//}
			sinceReseed = 0;
			if( worstRep != 0 ) {
				swapReplicas(0, worstRep);
			}
		}
#endif
		// determine if the exchange is odd or even
		int offset = exch;
		exch = (exch + 1) % 2;
		
		// we use the offset to vary the replica exchange sequence between the following
		// two patterns: (T1,T2),(T3,T4),(T5,T6)...(TN-1,TN) and (T2,T3),(T4,T5),..,(TN-2,TN-1)
		int cond = numReplicas - offset - 1;
		int i = 0 + offset;
		
		while( i < cond + offset ) {
			int j = i + 1;

#ifdef TRACE_REP_SWAPS
			//repSwaps[ tempMap[ replicas[i]->getTemperature() ] ][ tempMap[ replicas[j]->getTemperature() ] ][0]++;
			repSwaps[i][j][0]++;
#endif
			if( shouldSwapReplicas(i,j) ) {
#ifdef TRACE_REP_SWAPS
				//repSwaps[ tempMap[ replicas[i]->getTemperature() ] ][ tempMap[ replicas[j]->getTemperature() ] ][1]++;
				repSwaps[i][j][1]++;
#endif
				swapReplicas(i,j);
			}
			
			//if( repSwaps[i][j][0] < repSwaps[i][j][1] ) {
			//	cout << "something went awry!" << endl;
			//}
			
			i += 2;
		}
		
		
		if( bestSolution != prevBestSolution ) {
#ifdef TRACE_VERBOSE
			cout << timer.read() << ": Current Best Solution: " << bestSolution << endl;
#endif
			prevBestSolution = bestSolution;
		}
#if MOVESET == MIXED
		Conformation::swapOffset();
#endif
	}

	cout << "Real time: " << timer.read() << endl;
	timer.stop();
	
	
#if DIM == 2

cout << endl;
cout << bestRelDirections << endl << endl;

Conformation *c = new Conformation(params.sequence,200,"",1);
c->loadConformation(bestRelDirections);
c->printConformation();

delete c;
#endif	
	
#if defined TRACE_SQT_GENERAL_IMMEDIATE
	sqt.close();
#endif
	
#ifdef TRACE_REP_SWAPS	
{	
	ReplicaIterator replicaItr;
	
	for( replicaItr = replicas.begin(); replicaItr != replicas.end() - 1; replicaItr++ ) {
		Replica *rep1 = *replicaItr;
		Replica *rep2 = *(replicaItr + 1);
		double temp1 = rep1->getTemperature();
		double temp2 = rep2->getTemperature();
		int repId1 = tempMap[ temp1 ];
		int repId2 = tempMap[ temp2 ];
		
		cout << "(" << temp1 << "->" << temp2 << ")="
			 << (double)((double)repSwaps[repId1][repId2][1]/(double)repSwaps[repId1][repId2][0]) 
			 << endl;
	}
}
#endif

#ifdef TRACE_ARI_MOVES
{
	ostringstream ari_filename;
	ari_filename << "ari_moves_" << params.experimentID << ".txt";
	ofstream ari;
	ari.open(ari_filename.str().c_str(),ios::out);
	ari << setw(13) << "Temperature" << setw(13) << "Accept"
	    << setw(13) << "Reject" << setw(13) << "Impossible" 
	    << setw(13) << "Total"  << setw(16) << "Avg Num Res." << endl;
	
	ReplicaIterator replicaItr;
		
	
	for( replicaItr = replicas.begin(); replicaItr != replicas.end(); replicaItr++ ) {
		Replica *replica = *replicaItr;
		double temp = replica->getTemperature();
		int repId = tempMap[ temp ];
		
		ari << setprecision(4);
		ari << setw(13) << temp
			<< setprecision(2)
			<< setw(13) << ariMoves[repId][0] / ariMoves[repId][3]
			<< setw(13) << ariMoves[repId][1] / ariMoves[repId][3]
			<< setw(13) << ariMoves[repId][2] / ariMoves[repId][3]
			<< setprecision(10)
			<< setw(13) << ariMoves[repId][3]
			<< setprecision(3)
			<< setw(16) << ariMoves[repId][4] / ariMoves[repId][0] << endl;
	}
	ari.close();
	
	ostringstream eval_per_second_filename;
	eval_per_second_filename << "eps_" << params.experimentID << ".txt";
	ofstream eps;
	eps.open(eval_per_second_filename.str().c_str(),ios::out);
    eps << setprecision(4);
    eps << ariMoves[0][3] / timer.read() << endl;
    eps.close();
}	
#endif

#ifdef TRACE_MST
	for(int i=0; i<numReplicas; i++) {
		replicas[i]->writeResults();
	}
#endif
	
}
