/*
 *  Project: Replica exchange Monte Carlo for Protein Folding in the HP model
 *  File:    SearchResult.cpp
 *
 *  Note: The latest source code and information regarding this project 
 *  is available at http://www.cs.ubc.ca/labs/beta/Projects/REMC-HPPFP/
 *
 *  For citation purposes, please use:
 *  Thachuk C, Shmygelska A and Hoos H: Replica exchange Monte Carlo for 
 *  protein folding in the HP model. BMC Bioinformatics 2007, Submitted.
 * 
 *  Copyright (C) 2007  Chris Thachuk, Alena Shmygelska and Holger H Hoos
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */


#include "SearchResult.h"

#ifdef TRACE_MST
string SearchResult::getMSTrace() {
	
	ostringstream result;
	result << "RUNLENGTH\tRUNTIME\tFREE_ENERGY\tTEMPERATURE" << endl;
	
	vector<SQT_item>::const_iterator itr;
	
	for( itr=ms_trace.begin(); itr != ms_trace.end(); itr++ ) {
		result	<< itr->runLength << "\t"
				<< itr->runTime	<< "\t"
				<< itr->freeEnergy << "\t"
				<< itr->T << endl;
	}
	
	result << endl;
	
    return result.str();
}
#endif

SearchResult::SearchResult(const string &expID, const int repID) {
	setExperimentID(expID);
	setReplicaID(repID);
	
	movesAccepted = 0;
	movesRejected = 0;
	movesImpossible = 0;
	
	ms_sqt.runLength = 0;
	ms_sqt.runTime = 0.0;
	ms_sqt.freeEnergy = 0;
	ms_sqt.relDirections = "";
	
#ifdef TRACE_SQT_DETAIL
	ostringstream sqt_filename;
	sqt_filename << "sqt_" << experimentID << "_" << replicaID << ".txt";
	sqt.open(sqt_filename.str().c_str(), ios::out);
	sqt << "RUNLENGTH\tRUNTIME\tFREE_ENERGY\tTEMPERATURE" << endl;
#endif	
}

void SearchResult::setAcceptedMoves(long double num) {
	movesAccepted = num;
}

void SearchResult::setRejectedMoves(long double num) {
	movesRejected = num;
}

void SearchResult::setImpossibleMoves(long double num) {
	movesImpossible = num;
}

void SearchResult::setNumSteps(long double num) {
	numSteps = num;
}

void SearchResult::setNumResidues(long double num) {
	numResidues = num;
}

long double SearchResult::getAcceptedMoves() {
	return movesAccepted;
}

long double SearchResult::getRejectedMoves() {
	return movesRejected;
}

long double SearchResult::getImpossibleMoves() {
	return movesImpossible;
}

long double SearchResult::getNumSteps() {
	return numSteps;
}

long double SearchResult::getNumResidues() {
	return numResidues;
}

void SearchResult::writeResults() {
	// write out the best solution trace file
#ifdef TRACE_MST
	ostringstream mst_filename;
	mst_filename << "mst_" << experimentID << "_" << replicaID << ".txt";
		
	ofstream mst_out;
	mst_out.open(mst_filename.str().c_str(), ios::out);
	mst_out << getMSTrace() << endl;
	mst_out.flush();
	mst_out.close();
#endif

// finish write to solution quality trace file
#ifdef TRACE_SQT_DETAIL
	sqt.flush();
	sqt.close();
#endif
}

void SearchResult::setExperimentID(const string &expID) {
	experimentID = expID;
}

void SearchResult::setReplicaID(const int repID) {
	replicaID = repID;
}

string& SearchResult::getExperimentID() {
	return experimentID;
}

int SearchResult::getReplicaID() {
	return replicaID;
}

void SearchResult::addSQT_item( const long double run_length, const float run_time, 
	const int free_energy, const double T, const int relDirections[], const int dirLength ) {
	
	//SQT_item sqt_item;
	//sqt_item.runLength = run_length;
	//sqt_item.runTime = run_time;{
	//sqt_item.freeEnergy = free_energy;
	//sqt_item.T = T;
	//sqt_item.relDirections = "";
	
	if( free_energy < ms_sqt.freeEnergy ) {
		ms_sqt.runLength = run_length;
		ms_sqt.runTime = run_time;
		ms_sqt.freeEnergy = free_energy;
		ms_sqt.T = T;
		ostringstream buf;
		for(int i=0; i<dirLength; i++) {
			switch(relDirections[i]) {
				case straight: 	buf << "s"; break;
				case left:		buf << "l"; break;
				case right:		buf << "r"; break;
				case up:		buf << "u"; break;
				case down:		buf << "d"; break;
				case ground:	buf << "g"; break;
			}
		}
		ms_sqt.relDirections = buf.str();

#ifdef TRACE_MST
//		ms_trace.push_back( sqt_item );
#endif

	}

#ifdef TRACE_SQT_DETAIL
//	sqt << sqt_item.runLength << "\t"
//		<< sqt_item.runTime	<< "\t"
//		<< sqt_item.freeEnergy << "\t"
//		<< sqt_item.T << endl;
#endif

}

void SearchResult::resetMinimumSolution() {
	ms_sqt.runLength = 0;
	ms_sqt.runTime = 0.0;
	ms_sqt.freeEnergy = 0;
	//ms_sqt.relDirections = "";
}

void SearchResult::setMinimumSolution(const int relDirections[], const int freeEnergy, const int dirLength) {
	ms_sqt.runLength = 0;
	ms_sqt.runTime = 0;
	ms_sqt.freeEnergy = freeEnergy;
	ms_sqt.T = 0;
	ostringstream buf;
	for(int i=0; i<dirLength; i++) {
		switch(relDirections[i]) {
			case straight: 	buf << "s"; break;
			case left:		buf << "l"; break;
			case right:		buf << "r"; break;
			case up:		buf << "u"; break;
			case down:		buf << "d"; break;
			case ground:	buf << "g"; break;
		}
	}
	ms_sqt.relDirections = buf.str();
}

SQT_item& SearchResult::getMinimumSolution() {
	return ms_sqt;
}

