/*
 *  Project: Replica exchange Monte Carlo for Protein Folding in the HP model
 *  File:    main.cpp
 *
 *  Note: The latest source code and information regarding this project 
 *  is available at http://www.cs.ubc.ca/labs/beta/Projects/REMC-HPPFP/
 *
 *  For citation purposes, please use:
 *  Thachuk C, Shmygelska A and Hoos H: Replica exchange Monte Carlo for 
 *  protein folding in the HP model. BMC Bioinformatics 2007, Submitted.
 * 
 *  Copyright (C) 2007  Chris Thachuk, Alena Shmygelska and Holger H Hoos
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include <iostream>
#include <iomanip>
#include <algorithm>
#include <numeric>
#include <vector>
#include <ctime>
#include <fstream>
#include <map>
#include <getopt.h>
#include <string>
#include "REMSimulation.h"
#include "SearchParameters.h"
#include "SearchResult.h"
#include "Random.h"
#include "Conformation.h"
#include "Const.h"

using namespace std;

void check_options( SearchParameters &params, long int argc, char *argv[] );
void print_program_id();
void compareFolds( string, int );
void validateFold( string, string, int );

string compFoldsFile = "";
string coFilename = "";
string cfFilename = "";
string ufFilename = "";
string fold = "";
int numSets = 1;
bool validate = false;

int main (int argc, char *argv[]) {
	SearchParameters params;
	check_options( params, argc, argv );
	
	if( validate == true ) {
		validateFold( params.sequence, fold, params.knownEnergy );
		return 0;
	}

	if( compFoldsFile != "" ) {
		compareFolds(compFoldsFile, numSets);
		return 0;
	}
		
	if( params.suppliedSeed != 0 ) {
		setSeed( params.suppliedSeed );
	} else {
		setSeed( time(0) );
	}
	
	print_program_id();
			
	REMSimulation simulation( params );
	vector<SQT_item> bestSolutionTrace;
	
	cout << "Begin Simulation" << endl;
	simulation.runSimulation(bestSolutionTrace);
	cout << "End Simulation" << endl << endl;
	
	return 0;
}

void validateFold( string sequence, string fold, int energy ) {
	Conformation *c = new Conformation(sequence,200,"",1);
    c->loadConformation(fold);
	cout << c->getDirections() << endl;
	c->recalculateDirections();
	cout << c->getDirections() << endl;
	
	if( ! c->testLatticePositions() ) {
		cout << "Invalid fold" << endl << endl;
		c->printRelLatticePositions();
		return;
	}
	
	if( ! c->testFreeEnergy(energy) ) {
		cout << "Invalid energy" << endl << endl;
		c->printRelLatticePositions();
		return;
	}
	cout << "Valid fold and energy" << endl << endl;
}

void print_program_id () {
	#if DIM == 3
		#if MOVESET == LOCAL
			cout << "REMC-HP3D-LM" << endl << endl;
		#elif MOVESET == PULLMOVES
			cout << "REMC-HP3D-PM" << endl << endl;
		#elif MOVESET == MIXED
			cout << "REMC-HP3D-M" << endl << endl;
		#elif MOVESET == MIXED2
			cout << "REMC-HP3D-M2" << endl << endl;
		#endif
	#else	
		#if MOVESET == LOCAL
			cout << "REMC-HP2D-LM" << endl << endl;
		#elif MOVESET == PULLMOVES
			cout << "REMC-HP2D-PM" << endl << endl;
		#elif MOVESET == MIXED
			cout << "REMC-HP2D-M" << endl << endl;
		#elif MOVESET == MIXED2
			cout << "REMC-HP2D-M2" << endl << endl;
		#endif
	#endif
}

void check_options( SearchParameters &params, long int argc, char *argv[] ) 
{
  int opt;
  const string delims(", ");
  string line;
  string::size_type begIdx, endIdx;
  
  while (1) {
    int option_index = 0;

    static struct option long_options[] = {
      {"runs",       1, 0, '1'}, 
      {"seq",        1, 0, '2'}, 
      {"eng",        1, 0, '3'},  
      {"expID",      1, 0, '4'},
	  {"numReps",	 1, 0, '5'},
	  {"minT",		 1, 0, '6'},
	  {"maxT",		 1, 0, '7'},
	  {"numLocalSteps", 1, 0, '8'},
	  {"maxRunTime",	1, 0, '9'},
	  {"traceFile",		1, 0, 'a'},
	  {"Tvals",			1, 0, 'b'},
	  {"seed",          1, 0, 'c'},
	  {"mu",			1, 0, 'd'},
	  {"reseedAfter",	1, 0, 'e'},
	  {"reseedTimeout", 1, 0, 'f'},
	  {"compareFolds",  1, 0, 'g'},
	  {"coFile",        1, 0, 'h'},
	  {"cfFile",        1, 0, 'i'},
	  {"ufFile",		1, 0, 'j'},
	  {"numSets",       1, 0, 'k'},
	  {"pmWeight",      1, 0, 'l'},
	  {"fold",			1, 0, 'm'},
	  {"initConf",		1, 0, 'n'},
      
      {0,0, 0,  0 }
    };
    
    opt = getopt_long_only(argc, argv, "1:2:3:4:5:6:7:8:9:a:b:c:d:e:f:g:h:i:j:k:l:m:n:",
		       long_options, &option_index);
    if ( opt == -1 )
      break;
    
    switch ( opt ){
    case 0:
      printf("Error: Confused by input on long option...\n"), exit(1);
      break;
    
	case '1':
		params.numRuns = atoi(optarg);
		break;
	
	case '2':
		params.sequence = optarg;
		break;
	
	case '3':
		params.knownEnergy = -(atoi(optarg));
		break;
	
	case '4':
		params.experimentID = optarg;
		break;
	
	case '5':
		params.numReplicas = atoi(optarg);
		break;
		
	case '6':
		params.minTemperature = atof(optarg);
		break;
	
	case '7':
		params.maxTemperature = atof(optarg);
		break;
	
	case '8':
		params.numLocalSteps = atoi(optarg);
		break;
		
	case '9':
		params.maxRunTime = atoi(optarg);
		break;
		
	case 'a':
		params.traceFile = optarg;
		break;
		
	case 'b':
		line = optarg;
		begIdx = line.find_first_not_of(delims);
		
		while (begIdx != string::npos) {
			endIdx = line.find_first_of(delims, begIdx);
			
			if (endIdx == string::npos) {
				endIdx = line.length();
			}
			
			params.temperatureValues.push_back(atof(line.substr(begIdx,endIdx - begIdx).c_str()));
			begIdx = line.find_first_not_of(delims, endIdx);
		}

		break;
	
	case 'c':
		params.suppliedSeed = atoi(optarg);
		break;
    
	case 'd':
		params.mu = atof(optarg);
		break;
	
	case 'e':
		params.reseedAfter = atof(optarg);
		break;
	
	case 'f':
		params.reseedTimeout = atof(optarg);
		break;
		
	case 'g':
		compFoldsFile = optarg;
		break;
		
	case 'h':
		coFilename = optarg;
		break;
	
	case 'i':
		cfFilename = optarg;
		break;
		
	case 'j':
		ufFilename = optarg;
		break;
		
	case 'k':
		numSets = atoi(optarg);
		break;
		
	case 'l':
		params.pmWeight = atof(optarg);
		break;
	
	case 'm':
		fold = optarg;
		validate = true;
		break;
	
	case 'n':
		params.initConf = optarg;
		break;
		
    default:
      printf("Error: Confused by input...\n"), exit(1);
      
    }
  }
  
} 

void compareFolds( string foldsFilename, int numSets ) {
	map< string, string > foldMap;
	map< string, vector<int> > foldFreq;
	
	ifstream foldsFile( foldsFilename.c_str(), ios::in );
	Conformation *c;
	string sequence;
	string fold;
	string contactString;
	string symContactString;
	int numFolds = 0;
	int minHsa = 99999;
	int maxHsa = 0;
	int totHsa = 0;
	vector< double > hhCO;
	vector< vector< int > > freqMatrix;
	
	if (foldsFile.is_open())
	{
		// read in the sequence
		if (! foldsFile.eof() ) {
			getline(foldsFile,sequence);
		}
		
		freqMatrix.resize(sequence.length() );
		vector< vector< int > >::iterator itr;
		
		for(itr=freqMatrix.begin(); itr!=freqMatrix.end(); itr++) {
			itr->resize(sequence.length() );
			vector<int>::iterator itr2;
			
			for(itr2=itr->begin(); itr2!=itr->end(); itr2++) {
				*itr2 = 0;
			}
		}
		
		// read in each conformation
	    while (! foldsFile.eof() )
	    {
		  string str;
		  int setnum;
	      getline (foldsFile,str);
	
		  string buf;
		  stringstream ss(str);
		  ss >> setnum;
		  ss >> fold;
		  
	      if(! foldsFile.eof() ) {
			numFolds++;
			c = new Conformation(sequence,200,"",1);
		    c->loadConformation(fold);
			contactString = c->getContactString();
			symContactString = c->getSymmetricContactString();
			 
			if( foldMap.find(contactString) != foldMap.end() ) {
				foldFreq[contactString][setnum-1]++;
			} else if( foldMap.find(symContactString) != foldMap.end() ) {
				foldFreq[symContactString][setnum-1]++;
			} else {
				foldMap[contactString] = c->getDirections();
				foldFreq[contactString].resize(numSets,0);
				foldFreq[contactString][setnum-1]++;
			}
			
			hhCO.push_back(c->relativeHHContactOrder());
			int sa = c->solventAccessibility();
			if(sa<minHsa) minHsa = sa;
			if(sa>maxHsa) maxHsa = sa;
			totHsa+=sa;
			delete c;
		  }
		
	    }
	
		map< string, string >::iterator foldItr;
		vector< int >::iterator vItr;
		
		int foldNum = 1;
		for(foldItr=foldMap.begin(); foldItr!=foldMap.end(); foldItr++) {
			string contactStr = foldItr->first;
			string relDir = foldItr->second;
			c = new Conformation(sequence,200,"",1);
			c->loadConformation(relDir);
			
			cout << "Unique Conformation " << foldNum++ << " (";
			bool first = true;
			int tot = 0;
			for(vItr=foldFreq[contactStr].begin(); vItr!=foldFreq[contactStr].end(); vItr++) {
				if(!first) cout << " + ";
				cout << *vItr;
				tot += *vItr;
				first = false;
			}
			cout << " = " << tot << ") of " << numFolds << endl;
			cout << "-----------------------------------------------" << endl;
			c->computeFreeEnergy();
			cout << "Free Energy              : " << c->getFreeEnergy() << endl;
			cout << "Relative HH CO           : " << c->relativeHHContactOrder() << endl;
			cout << "Absolute HH CO           : " << c->absoluteHHContactOrder() << endl;
			cout << "H solvent accessible area: " << c->solventAccessibility() << endl;
			//cout << "Contact String: " << c->getContactString() << endl;
			//cout << "SContact String:" << c->getSymmetricContactString() << endl;
#if DIM == 2
			c->printConformation();
#endif
			cout << endl;
			
			for(vItr=foldFreq[contactStr].begin(); vItr!=foldFreq[contactStr].end(); vItr++) {
				c->addToContactFrequencyMatrix( freqMatrix, *vItr );
			}
			delete c;
			
		}
	
		sort(hhCO.begin(),hhCO.end());
		
		cout << fixed << setprecision(4);
		cout << endl << "Unique conformations   : " << foldMap.size() << " of " << numFolds << endl << endl;
		cout << "Minimum relative HH CO : " << *(min_element(hhCO.begin(), hhCO.end())) << endl;
		cout << "Maximum relative HH CO : " << *(max_element(hhCO.begin(), hhCO.end())) << endl;
		cout << "Mean relative HH CO    : " << accumulate(hhCO.begin(), hhCO.end(),0.0) / hhCO.size() << endl << endl;
		cout << "Minimum H solvent aa   : " << minHsa << endl;
		cout << "Maximum H solvent aa   : " << maxHsa << endl;
		cout << "Mean H solvent aa      : " << (double)totHsa / numFolds << endl;

	    foldsFile.close();
		cout << endl;
		
		// do we write out the unique folds file ?
		if( ufFilename != "" ) {
			ofstream ufFile( ufFilename.c_str(), ios::out );
			if( ! ufFile.is_open() ) {
				cerr << "Unable to create file: "  << ufFilename << endl;
				return;
			}
			
			for(foldItr=foldMap.begin(); foldItr!=foldMap.end(); foldItr++) {
				string contactStr = foldItr->first;
				
				for(vItr=foldFreq[contactStr].begin(); vItr!=foldFreq[contactStr].end(); vItr++) {
					ufFile << *vItr << "\t";
				}
				
				ufFile << contactStr << endl;
			}
			
			ufFile.close();
		}
		
		// do we write out the contact frequency file ?
		if( cfFilename != "" ) {
			ofstream cfFile( cfFilename.c_str(), ios::out );
			if( ! cfFile.is_open() ) {
				cerr << "Unable to create file: "  << cfFilename << endl;
				return;
			}
			cfFile << fixed << setprecision(2);

			//vector< vector< int > >::iterator itr;
			int row=1;
			cfFile << setw(6) << " ";
			for(itr=freqMatrix.begin(); itr!=freqMatrix.end(); itr++, row++) {
				if( sequence[row-1] != 'h' ) continue;
				cfFile << setw(6) << row;
			}
			cfFile << endl; row = 1;

			for(itr=freqMatrix.begin(); itr!=freqMatrix.end(); itr++, row++) {
				if( sequence[row-1] != 'h' ) continue;

				cfFile << setw(6) << row;

				vector<int>::iterator itr2;
				int col=1;
				for(itr2=itr->begin(); itr2!=itr->end(); itr2++, col++) {
					if( sequence[col-1] != 'h' ) continue;

					//if( col < row ) {
					//	cfFile << setw(4) << " ";
					//} else {
						cfFile << setw(6) << *itr2;
					//}
				}
				cfFile << endl;
			}
			cfFile.close();
		}
		
		// do we write out the contact order file ?
		if( coFilename != "" ) {
			ofstream coFile( coFilename.c_str(), ios::out );
			if( ! coFile.is_open() ) {
				cerr << "Unable to create file: "  << coFilename << endl;
				return;
			}
			
			int confNum = 1;
			vector<double>::iterator itr;
			double prevCO = hhCO[0];
			int bucket = 1;
			int total = 1;
			
			for(itr=hhCO.begin() + 1; itr!=hhCO.end(); itr++) {
				if( *itr != prevCO ) {
					coFile << prevCO << "\t" << total << "\t" << bucket << endl;
					bucket = 1;
					prevCO = *itr;
				} else {
					bucket++;
				}
				total++;
			}
			coFile << prevCO << "\t" << total << "\t" << bucket << endl;
			coFile.close();
		}
		
	} else {
		cerr << "Unable to open file: " << foldsFilename << endl;
		return;
	}	
}

